# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from europi import *
from europi_script import EuroPiScript

from time import sleep, sleep_ms, ticks_ms, ticks_diff
from math import log


#Output a 10ms trigger to indicate end-of-rise and end-of-fall on CV5/6
TRIGGER_DURATION_MS = 10

#Constants for tracking the direction of the envelope's voltage
DIRECTION_RISING = 1
DIRECTION_FALLING = 0
DIRECTION_SUSTAIN = 3

#Sustain modes
SUSTAIN_MODE_AR = 0
SUSTAIN_MODE_ASR = 1

#Looping modes
LOOPING_MODE_LOOP = 1
LOOPING_MODE_ONCE = 0


class EnvelopeGenerator(EuroPiScript):
    def __init__(self):
        super().__init__()
        state = self.load_state_json()

        self.sustain_mode = state.get("sustain_mode", SUSTAIN_MODE_ASR)
        self.looping_mode = state.get("looping_mode", LOOPING_MODE_ONCE)

        #Milliscond tick of of the most recent end-of-rise and end-of-fall
        #initialized to be 2x trigger duration in the past on startup to prevent roll-over issues
        self.last_rise_end_at = time.ticks_add(time.ticks_ms(), -2*TRIGGER_DURATION_MS)
        self.last_fall_end_at = self.last_rise_end_at

        self.max_output_voltage = europi_config.MAX_OUTPUT_VOLTAGE

        #Distance of envelope voltage from max voltage/0 before 'jumping' to it - prevents large logarithmic calculations
        self.voltage_threshold = 0.1

        #Length of the longest possible envelope (not in any meaningful unit)
        self.max_increment_factor = 256
        self.update_increment_factor()

        #Time in ms between incrementing value of envelope
        self.increment_delay = 1

        #0 will start a new envelope at the current value, 1 will start it from zero
        self.retrig_mode = 0

        #Display refresh rate in ms
        self.display_refresh_rate = 30
        self.last_refreshed_display = self.display_refresh_rate

        self.envelope_display_bounds = [0, 0, int(OLED_WIDTH), int(OLED_HEIGHT / 2)]

        self.direction = DIRECTION_SUSTAIN  # DIRECTION_RISING, DIRECTION_FALLING, or DIRECTION_SUSTAIN
        self.envelope_value = 0

        self.envelope_out = cv2
        self.envelope_inverted_out = cv3
        self.din_copy_out = cv1

        self.sustain_gate = cv4
        self.eor_trigger = cv5
        self.eof_trigger = cv6

        din.handler(self.receive_trigger_rise)
        din.handler_falling(self.receive_trigger_fall)
        b1.handler(self.change_sustain_mode)
        b2.handler(self.change_looping_mode)


    @classmethod
    def display_name(cls):
        return "EnvelopeGen"

    def receive_trigger_rise(self):
        if self.retrig_mode == 1:
            self.envelope_value = 0
        self.direction = 1

    def receive_trigger_fall(self):
        if self.direction == DIRECTION_RISING:
            # Interrupted rise; output the trigger because we've risen as high as we're going to
            self.last_rise_end_at = time.ticks_ms()
        self.direction = DIRECTION_FALLING

    def change_sustain_mode(self):
        self.sustain_mode = 1 - self.sustain_mode
        #Save state to file
        self.save_state()

    def change_looping_mode(self):
        self.looping_mode = 1 - self.looping_mode
        #Save state to file
        self.save_state()

    def copy_digital_input(self):
        self.din_copy_out.value(din.value())

    def difference(self, a, b):
        return abs(a - b)

    def update_increment_factor(self):
        increment_factor_rising = k1.range(self.max_increment_factor, 512) / 2
        self.increment_factor = [(increment_factor_rising + 1), (k2.range(self.max_increment_factor, 256) + 1 + (ain.percent(256) * self.max_increment_factor))]

    def log(self, number):
        return log(max(number, 1))

    def update_envelope_value(self):
        #Envelope rising
        if self.direction == DIRECTION_RISING:
            increment = self.difference(self.envelope_value, self.max_output_voltage) / self.increment_factor[0]
            self.envelope_value += increment
            if self.difference(self.envelope_value, self.max_output_voltage) <= self.voltage_threshold:
                self.envelope_value = self.max_output_voltage
                if self.sustain_mode == 1 and self.looping_mode == 0:
                    self.direction = DIRECTION_SUSTAIN
                else:
                    self.direction = DIRECTION_FALLING
                self.last_rise_end_at = time.ticks_ms()
            else:
                sleep_ms(self.increment_delay)

        #Envelope falling
        elif self.direction == DIRECTION_FALLING:
            increment = self.difference(0, self.envelope_value) / self.increment_factor[1]
            self.envelope_value -= increment
            if self.difference(0, self.envelope_value) <= self.voltage_threshold:
                self.envelope_value = 0
                if self.looping_mode == LOOPING_MODE_ONCE:
                    self.direction = DIRECTION_SUSTAIN
                else:
                    self.direction = DIRECTION_RISING
                self.last_fall_end_at = time.ticks_ms()
            else:
                sleep_ms(self.increment_delay)

        #Update CV output to envelope value
        self.update_output_voltage()


    def update_output_voltage(self):
        self.envelope_out.voltage(self.envelope_value)
        self.envelope_inverted_out.voltage(self.max_output_voltage - self.envelope_value)

        if self.direction == DIRECTION_SUSTAIN:
            self.sustain_gate.on()
        else:
            self.sustain_gate.off()

        now = time.ticks_ms()
        if time.ticks_diff(now, self.last_rise_end_at) <= TRIGGER_DURATION_MS:
            self.eor_trigger.on()
        else:
            self.eor_trigger.off()

        if time.ticks_diff(now, self.last_fall_end_at) <= TRIGGER_DURATION_MS:
            self.eof_trigger.on()
        else:
            self.eof_trigger.off()

    def update_display(self):
        if ticks_diff(ticks_ms(), self.last_refreshed_display) >= self.display_refresh_rate:

            #Draw slope graph axis
            oled.hline(self.envelope_display_bounds[0], self.envelope_display_bounds[3], self.envelope_display_bounds[2], 1)
            oled.vline(self.envelope_display_bounds[0], self.envelope_display_bounds[1], self.envelope_display_bounds[3], 1)
            oled.vline((self.envelope_display_bounds[2] - 1), self.envelope_display_bounds[1], self.envelope_display_bounds[3], 1)

            try:
                rise_width = (self.increment_factor[0] - 1) / ((self.increment_factor[0] - 1) + (self.increment_factor[1] - 1))	#If envelope has zero rise and zero fall this will throw a ZeroDivisonError
                draw_envelope = True
            except ZeroDivisionError:
                draw_envelope = False

            if draw_envelope == True:
                rise_width_pixels = int(rise_width * self.envelope_display_bounds[2])
                fall_width = 1 - rise_width
                fall_width_pixels = int(self.envelope_display_bounds[2] - rise_width_pixels)

                #Generate rise slope pixels
                rise_pixels = []
                for pixel in range(rise_width_pixels):
                    x = pixel / (rise_width_pixels + 1)
                    y = x**2
                    x_pixel = rise_width_pixels - int(x * rise_width_pixels)
                    y_pixel = int(y * (self.envelope_display_bounds[3] - self.envelope_display_bounds[1]))
                    rise_pixels.append((x_pixel, y_pixel))
                rise_pixels.append((self.envelope_display_bounds[0], self.envelope_display_bounds[3]))

                #Generate fall slope pixels
                fall_pixels = []
                for pixel in range(fall_width_pixels):
                    x = pixel / (fall_width_pixels + 1)
                    y = x**2
                    x_pixel = (fall_width_pixels - int(x * fall_width_pixels)) + rise_width_pixels
                    y_pixel = self.envelope_display_bounds[3] - int(y * (self.envelope_display_bounds[3] - self.envelope_display_bounds[1]))
                    fall_pixels.append((x_pixel, y_pixel))
                fall_pixels.append((rise_width_pixels, self.envelope_display_bounds[1]))

                #Draw rise and fall slopes
                for array in [rise_pixels, fall_pixels]:
                    for index, pixel in enumerate(array[:-1]):
                        oled.line(array[index + 1][0], array[index + 1][1], pixel[0], pixel[1], 1)

                #Draw current envelope position
                current_envelope_position = 0
                if self.direction == DIRECTION_RISING or self.direction == DIRECTION_SUSTAIN:
                    current_envelope_position = int((self.envelope_value / self.max_output_voltage) * rise_width_pixels)
                elif self.direction == DIRECTION_FALLING:
                    current_envelope_position = self.envelope_display_bounds[2] - 1 - int((self.envelope_value / self.max_output_voltage) * (self.envelope_display_bounds[2] - rise_width_pixels))
                oled.vline(current_envelope_position, self.envelope_display_bounds[1], (self.envelope_display_bounds[3] - 1), 1)
            else:
                oled.vline(self.envelope_display_bounds[0], self.envelope_display_bounds[1], self.envelope_display_bounds[3], 1)
                oled.hline(self.envelope_display_bounds[0], self.envelope_display_bounds[1], self.envelope_display_bounds[2], 1)
                oled.vline((self.envelope_display_bounds[2] - 1), self.envelope_display_bounds[1], self.envelope_display_bounds[3], 1)

            #Display current envelope direction
            if self.direction == DIRECTION_RISING:
                direction_text = 'rise'
            elif self.direction == DIRECTION_FALLING:
                direction_text = 'fall'
            elif self.direction == DIRECTION_SUSTAIN and self.envelope_value == self.max_output_voltage:
                direction_text = 'hold'
            else:
                direction_text = 'off'
            oled.text(direction_text, 0, 20, 1)

            #Display current envelope mode (AR or ASR)
            if self.sustain_mode == SUSTAIN_MODE_AR:
                sustain_mode_text = 'ar'
            else:
                sustain_mode_text = 'asr'
            oled.text(sustain_mode_text, 50 + (4 if sustain_mode_text == 'ar' else 0), 20, 1)

            #Display current envelope looping mode
            if self.looping_mode == LOOPING_MODE_ONCE:
                looping_mode_text = 'once'
            else:
                looping_mode_text = 'loop'
            oled.text(looping_mode_text, 94, 20, 1)

            oled.show()
            oled.fill(0)
            self.last_refreshed_display = ticks_ms()

    def save_state(self):
        """Save the current state variables as JSON."""
        # Don't save if it has been less than 5 seconds since last save.
        if self.last_saved() < 5000:
            return

        state = {
            "sustain_mode": self.sustain_mode,
            "looping_mode": self.looping_mode,
        }
        self.save_state_json(state)

    def main(self):
        while True:
            self.copy_digital_input()
            self.update_display()
            self.update_increment_factor()
            self.update_envelope_value()

if __name__ == "__main__":
    EnvelopeGenerator().main()



# Copyright 2025 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Interface program for using Open Sound Control to set CV levels

The following addresses & types are used:

- /europi/cv{1-6} float: set raw voltage (0-MAX_OUTPUT_VOLTAGE)
- /europi/cv{1-6} int: set gate on/off (treated as boolean)
- /europi/cvs : as above, but allows all outputs to be set at once

The application settings can be used to change the root namespace
"""

from europi import *
from europi_script import EuroPiScript

import configuration
from machine import Timer

from experimental.osc import *
import experimental.wifi

class OscControl(EuroPiScript):
    def __init__(self):
        super().__init__()

        self.sanitize_osc_config()

        self.server = OpenSoundServer(
            recv_port=self.config.RECV_PORT,
            send_port=self.config.SEND_PORT,
            send_addr=self.config.SEND_ADDR,
        )

        self.ui_dirty = False

        @self.server.data_handler
        def on_data_recv(connection=None, data=None):
            if data.address == self.cv1_topic:
                self.set_cv(cv1, data)
            elif data.address == self.cv2_topic:
                self.set_cv(cv2, data)
            elif data.address == self.cv3_topic:
                self.set_cv(cv3, data)
            elif data.address == self.cv4_topic:
                self.set_cv(cv4, data)
            elif data.address == self.cv5_topic:
                self.set_cv(cv5, data)
            elif data.address == self.cv6_topic:
                self.set_cv(cv6, data)
            elif data.address == self.cvs_topic:
                # set all CVs at once
                self.set_cvs(data)

    def sanitize_osc_config(self):
        self.namespace = self.config.NAMESPACE
        if not self.namespace.startswith("/"):
            self.namespace = f"/{self.namespace}"
        if not self.namespace.endswith("/"):
            self.namespace = f"{self.namespace}/"

        def sanitize_topic(param_name):
            topic = ""
            if self.config[param_name].startswith("/"):
                topic = self.config[param_name]
            else:
                topic = f"{self.namespace}{self.config[param_name]}"
            topic = topic.rstrip("/")
            return topic

        self.cv1_topic = sanitize_topic("CV1_TOPIC")
        self.cv2_topic = sanitize_topic("CV2_TOPIC")
        self.cv3_topic = sanitize_topic("CV3_TOPIC")
        self.cv4_topic = sanitize_topic("CV4_TOPIC")
        self.cv5_topic = sanitize_topic("CV5_TOPIC")
        self.cv6_topic = sanitize_topic("CV6_TOPIC")
        self.cvs_topic = sanitize_topic("CVS_TOPIC")
        self.ain_topic = sanitize_topic("AIN_TOPIC")
        self.din_topic = sanitize_topic("DIN_TOPIC")
        self.k1_topic = sanitize_topic("K1_TOPIC")
        self.k2_topic = sanitize_topic("K2_TOPIC")
        self.b1_topic = sanitize_topic("B1_TOPIC")
        self.b2_topic = sanitize_topic("B2_TOPIC")

    def set_cv(self, cv_out, data):
        """
        Set the level for a single CV

        @param cv_out  CV1-6
        @param data  The OpenSoundPacket we're processing
        """
        if type(data.values[0]) is int:
            if data.values[0] == 0:
                cv_out.off()
            else:
                cv_out.on()
        elif type(data.values[0]) is float:
            cv_out.voltage(data.values[0] * europi_config.MAX_OUTPUT_VOLTAGE)
        self.ui_dirty = True

    def set_cvs(data):
        """
        Set the level for all CVs

        @param data  The OpenSoundPacket we're processing
        """
        for i in range(len(cvs)):
            v = data.values[i]
            t = type(data.values[i])
            cv = cvs[i]

            if t is int:
                if v == 0:
                    cv.off()
                else:
                    cv.on()
            elif t is float:
                cv.voltage(v * europi_config.MAX_OUTPUT_VOLTAGE)
        self.ui_dirty = True

    @classmethod
    def config_points(cls):
        return [
            # OSC namespace & topics
            configuration.string("NAMESPACE", default="/europi/"),
            configuration.string("CV1_TOPIC", default="cv1"),
            configuration.string("CV2_TOPIC", default="cv2"),
            configuration.string("CV3_TOPIC", default="cv3"),
            configuration.string("CV4_TOPIC", default="cv4"),
            configuration.string("CV5_TOPIC", default="cv5"),
            configuration.string("CV6_TOPIC", default="cv6"),
            configuration.string("CVS_TOPIC", default="cvs"),
            configuration.string("AIN_TOPIC", default="ain"),
            configuration.string("DIN_TOPIC", default="din"),
            configuration.string("K1_TOPIC", default="k1"),
            configuration.string("K2_TOPIC", default="k2"),
            configuration.string("B1_TOPIC", default="b1"),
            configuration.string("B2_TOPIC", default="b2"),

            # Network settings
            configuration.integer("RECV_PORT", default=9000, minimum=0, maximum=65535),
            configuration.integer("SEND_PORT", default=9001, minimum=0, maximum=65535),
            configuration.string("SEND_ADDR", default="192.168.4.100"),
        ]

    def on_status_tick(self, timer):
        # send the states of all of our inputs
        self.server.send_data(
            self.k1_topic,
            k1.percent(),
        )
        self.server.send_data(
            self.k2_topic,
            k2.percent(),
        )
        self.server.send_data(
            self.ain_topic,
            ain.percent(),
        )
        self.server.send_data(
            self.b1_topic,
            b1.value() != 0,
        )
        self.server.send_data(
            self.b2_topic,
            b2.value() != 0,
        )
        self.server.send_data(
            self.din_topic,
            din.value() != 0,
        )

    def draw(self):
        oled.fill(0)
        width = OLED_WIDTH // 6 -2
        for i in range(len(cvs)):
            cv = cvs[i]
            left = i * OLED_WIDTH // 6
            height = int(max(OLED_HEIGHT * cv.voltage() / europi_config.MAX_OUTPUT_VOLTAGE, 1))

            oled.fill_rect(left + 1, OLED_HEIGHT - height, width, height, 1)
        oled.show()
        self.ui_dirty = False

    def main(self):
        if wifi_connection is None:
            raise experimental.wifi.WifiError("No wifi connection")

        oled.centre_text(f"""{wifi_connection.ip_addr}
waiting...""")

        # start a timer to send our input state at 10Hz
        timer = Timer()
        timer.init(freq=20, mode=Timer.PERIODIC, callback=self.on_status_tick)

        while True:
            self.server.receive_data()
            if self.ui_dirty:
                self.draw()


if __name__ == "__main__":
    OscControl().main()



#!/usr/bin/env python3
# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A 1D physics simulating LFO, inspired by the ADDAC503
(https://www.addacsystem.com/en/products/modules/addac500-series/addac503)
"""

from europi import *
from europi_script import EuroPiScript

from experimental.knobs import KnobBank
from experimental.math_extras import rescale

import math
import time


EARTH_GRAVITY = 9.8
MIN_GRAVITY = 0.1
MAX_GRAVITY = 20

MIN_HEIGHT = 0.1
MAX_HEIGHT = 10.0

MIN_SPEED = -5.0
MAX_SPEED = 5.0

MIN_ELASTICITY = 0.0
MAX_ELASTICITY = 0.9

## If a bounce reaches no higher than this, assume we've come to rest
ASSUME_STOP_PEAK = 0.002


class Particle:
    def __init__(self):
        self.y = 0.0
        self.dy = 0.0

        self.last_update_at = time.ticks_ms()

        self.hit_ground = False
        self.reached_apogee = False
        self.stopped = True

        self.peak_height = 0.0

    def set_initial_position(self, height, velocity):
        self.peak_height = height
        self.y = height
        self.dy = velocity
        self.last_update_at = time.ticks_ms()

    def update(self, g, elasticity):
        """Update the particle position based on the ambient gravity & elasticy of the particle
        """
        now = time.ticks_ms()
        delta_t = time.ticks_diff(now, self.last_update_at) / 1000.0

        new_dy = self.dy - delta_t * g
        new_y = self.y + self.dy * delta_t

        # if we were going up, but now we're going down we've reached apogee
        self.reached_apogee = new_dy <= 0 and self.dy >= 0

        if self.reached_apogee:
            self.peak_height = self.y

        # if the vertical position is zero or negative, we've hit the ground
        self.hit_ground = new_y <= 0

        if self.hit_ground:
            #new_y = 0
            new_dy = abs(self.dy * elasticity)  # bounce upwards, reduding the velocity by our elasticity modifier

        self.stopped = self.peak_height <= ASSUME_STOP_PEAK

        if self.stopped:
            new_y = 0
            new_dy = 0

        self.dy = new_dy
        self.y = new_y
        self.last_update_at = now

class ParticlePhysics(EuroPiScript):
    def __init__(self):
        settings = self.load_state_json()

        self.gravity = settings.get("gravity", 9.8)
        self.initial_velocity = settings.get("initial_velocity", 0.0)
        self.release_height = settings.get("height", 10.0)
        self.elasticity = settings.get("elasticity", 0.75)

        self.k1_bank =  (
            KnobBank.builder(k1)
            .with_locked_knob("height", initial_percentage_value=rescale(self.release_height, MIN_HEIGHT, MAX_HEIGHT, 0, 1))
            .with_locked_knob("gravity", initial_percentage_value=rescale(self.gravity, MIN_GRAVITY, MAX_GRAVITY, 0, 1))
            .build()
        )

        self.k2_bank =  (
            KnobBank.builder(k2)
            .with_locked_knob("elasticity", initial_percentage_value=rescale(self.elasticity, MIN_ELASTICITY, MAX_ELASTICITY, 0, 1))
            .with_locked_knob("speed", initial_percentage_value=rescale(self.initial_velocity, MIN_SPEED, MAX_SPEED, 0, 1))

            .build()
        )

        self.particle = Particle()

        self.release_before_next_update = False

        self.alt_knobs = False

        @din.handler
        def on_din_rising():
            self.reset()

        @b1.handler
        def on_b1_press():
            self.reset()

        @b2.handler
        def on_b2_press():
            self.alt_knobs = True
            self.k1_bank.next()
            self.k2_bank.next()

        @b2.handler_falling
        def on_b2_release():
            self.alt_knobs = False
            self.k1_bank.next()
            self.k2_bank.next()


    @classmethod
    def display_name(cls):
        return "ParticlePhysics"

    def save(self):
        state = {
            "gravity"          : self.gravity,
            "initial_velocity" : self.initial_velocity,
            "height"           : self.release_height,
            "elasticity"       : self.elasticity
        }
        self.save_state_json(state)

    def reset(self):
        self.release_before_next_update = True

    def draw(self):
        oled.fill(0)
        row_1_color = 1
        row_2_color = 2
        if self.alt_knobs:
            oled.fill_rect(0, CHAR_HEIGHT+1, OLED_WIDTH, CHAR_HEIGHT+1, 1)
            row_2_color = 0
        else:
            oled.fill_rect(0, 0, OLED_WIDTH, CHAR_HEIGHT+1, 1)
            row_1_color = 0


        oled.text(f"h: {self.release_height:0.2f}  e: {self.elasticity:0.2f}", 0, 0, row_1_color)
        oled.text(f"g: {self.gravity:0.2f}  v: {self.initial_velocity:0.2f}", 0, CHAR_HEIGHT+1, row_2_color)

        # a horizontal representation of the particle bouncing off the left edge of the screen
        oled.pixel(int(rescale(self.particle.y, 0, self.release_height, 0, OLED_WIDTH)), 3 * CHAR_HEIGHT, 1)

        oled.show()

    def main(self):
        while True:
            g = round(rescale(self.k1_bank["gravity"].percent(), 0, 1, MIN_GRAVITY, MAX_GRAVITY), 2)
            h = round(rescale(self.k1_bank["height"].percent(), 0, 1, MIN_HEIGHT, MAX_HEIGHT), 2)
            v = round(rescale(self.k2_bank["speed"].percent(), 0, 1, MIN_SPEED, MAX_SPEED), 2)
            e = round(rescale(self.k2_bank["elasticity"].percent(), 0, 1, MIN_ELASTICITY, MAX_ELASTICITY), 2)

            # the maximum veliocity we can attain, given the current parameters
            # d = 1/2 aT^2 -> T = sqrt(2d/a)
            h2 = 0
            v2 = 0
            if v > 0:
                # initial upward velocity; add this to the initial height
                t = v / g
                h2 = v * t
            else:
                v2 = abs(v)
            t = math.sqrt(2 * (h+h2) / g)
            max_v = g * t + v2

            if g != self.gravity or \
                    h != self.release_height or \
                    v != self.initial_velocity or \
                    e != self.elasticity:
                self.gravity = g
                self.initial_velocity = v
                self.release_height = h
                self.elasticity = e
                self.save()

            self.draw()

            if self.release_before_next_update:
                self.particle.set_initial_position(self.release_height, self.initial_velocity)
                self.release_before_next_update = False

            self.particle.update(self.gravity, self.elasticity)

            # CV 1 outputs a gate whenever we hit the ground
            if self.particle.hit_ground:
                cv1.on()
            else:
                cv1.off()

            # CV 2 outputs a trigger whenever we reach peak altitude and start falling again
            if self.particle.reached_apogee:
                cv2.on()
            else:
                cv2.off()

            # CV 3 outputs a gate when the particle comes to rest
            if self.particle.stopped:
                cv3.on()
            else:
                cv3.off()

            # CV 4 outputs control voltage based on the height of the particle
            cv4.voltage(rescale(self.particle.y, 0, MAX_HEIGHT, 0, MAX_OUTPUT_VOLTAGE))

            # CV 5 outputs control voltage based on the speed of the particle
            cv5.voltage(rescale(abs(self.particle.dy), 0, max_v, 0, MAX_OUTPUT_VOLTAGE))

            # TODO: I don't know what to use CV6 for. But hopefully I'll think of something
            cv6.off()


if __name__ == "__main__":
    ParticlePhysics().main()



# Copyright 2025 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Serves a simple HTTP page to control the levels of the six CV outputs from a browser

Requires a Pico W or Pico 2 W with a valid wifi setup to work
"""

from europi import *
from europi_script import EuroPiScript

from experimental.http_server import *
import json


HTML_DOCUMENT = """<!DOCTYPE html>
<html lang="en">
    <head>
        <style>
            body {
                font-family: Montserrat;
                text-align: center;
            }
            h1 {
                font-weight: normal;
                font-size: 2.5rem;
                letter-spacing: 1.75rem;
                padding-left: 1.75rem;
                text-align: center;
            }
            h2 {
                font-weight:  bold;
                font-size: 3.0rem;
            }
            td {
                font-weight: lighter;
                font-size: 2.0rem;
            }
            .content-wrapper {
                margin: 0;
                position: absolute;
                top: 50%;
                align-content: center;
                width: 100%;
                -ms-transform: translateY(-50%);
                transform: translateY(-50%);
            }
            table {
                margin: auto;
            }
        </style>
        <title>
            EuroPi Web Control
        </title>
        <script>
            function on_change() {
                cvs = {
                    "cv1": parseFloat(document.getElementById("cv1").value),
                    "cv2": parseFloat(document.getElementById("cv2").value),
                    "cv3": parseFloat(document.getElementById("cv3").value),
                    "cv4": parseFloat(document.getElementById("cv4").value),
                    "cv5": parseFloat(document.getElementById("cv5").value),
                    "cv6": parseFloat(document.getElementById("cv6").value)
                }
                console.debug(cvs)

                var xhr = new XMLHttpRequest();
                var url = document.URL;
                xhr.open("POST", url, true);
                xhr.setRequestHeader("Content-Type", "text/json");
                xhr.onreadystatechange = function () {
                    if (xhr.readyState === 4 && xhr.status === 200) {
                        var json = JSON.parse(xhr.responseText);
                        console.log(json);
                    }
                };
                var data = JSON.stringify(cvs);
                xhr.send(data);
            }
        </script>
    </head>
    <body>
        <h1>EuroPi Web Control</h1>
        <div class="content-wrapper">
            <table>
                <tr>
                    <td>
                        CV1
                    </td>
                    <td>
                        CV2
                    </td>
                    <td>
                        CV3
                    </td>
                </tr>
                <tr>
                    <td>
                        <input type="range" min="0" max="10" step="0.001" value="0" id="cv1" oninput="on_change()">
                    </td>
                    <td>
                        <input type="range" min="0" max="10" step="0.001" value="0" id="cv2" oninput="on_change()">
                    </td>
                    <td>
                        <input type="range" min="0" max="10" step="0.001" value="0" id="cv3" oninput="on_change()">
                    </td>
                </tr>
                <tr>
                    <td>
                        CV4
                    </td>
                    <td>
                        CV5
                    </td>
                    <td>
                        CV6
                    </td>
                </tr>
                <tr>
                    <td>
                        <input type="range" min="0" max="10" step="0.001" value="0" id="cv4" oninput="on_change()">
                    </td>
                    <td>
                        <input type="range" min="0" max="10" step="0.001" value="0" id="cv5" oninput="on_change()">
                    </td>
                    <td>
                        <input type="range" min="0" max="10" step="0.001" value="0" id="cv6" oninput="on_change()">
                    </td>
                </tr>
            </table>
        </div>
    </body>
</html>"""


class HttpControl(EuroPiScript):

    def __init__(self):
        super().__init__()

        self.server = HttpServer(80)

        @self.server.get_handler
        def handle_get(connection=None, request=None):
            self.server.send_html(
                connection,
                HTML_DOCUMENT,
            )

        @self.server.post_handler
        def handle_post(connection=None, request=None):
            """
            Process a POST request from the client

            The request should look like

                POST / HTTP/1.1
                Host: 10.0.0.167
                User-Agent: Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:135.0) Gecko/20100101 Firefox/135.0
                Accept: */*
                Accept-Language: en-US,en;q=0.5
                Accept-Encoding: gzip, deflate
                Content-Type: text/json
                Content-Length: 63
                Origin: http://10.0.0.167
                Connection: keep-alive
                Referer: http://10.0.0.167/

                {"cv1":7.071,"cv2":8.5,"cv3":0,"cv4":0,"cv5":3.857,"cv6":5.929}

            We only care about the actual JSON body
            """
            try:
                (_, body) = request.split("\r\n\r\n")
                jdata = json.loads(body)

                for i in range(NUM_CVS):
                    cvs[i].voltage(jdata.get(f"cv{i+1}", 0.0))

                self.server.send_json(
                    connection,
                    {
                        "inputs": {
                            "ain": ain.read_voltage(),
                            "din": din.value(),
                            "k1": k1.percent(),
                            "k2": k2.percent(),
                            "b1": b1.value(),
                            "b2": b2.value(),
                        },
                        "outputs": {
                            "cv1": cv1.voltage(),
                            "cv2": cv2.voltage(),
                            "cv3": cv3.voltage(),
                            "cv4": cv4.voltage(),
                            "cv5": cv5.voltage(),
                            "cv6": cv6.voltage(),
                        },
                    },
                    headers=None,
                )
            except ValueError as err:
                log_warn(f"{body} is not valid json")
                self.server.send_error_page(
                    err,
                    connection,
                    status=HttpStatus.BAD_REQUEST,
                )

    def main(self):
        if wifi_connection is None:
            raise WifiError("No wifi connection")

        while not wifi_connection.is_connected:
            oled.centre_text(f"""{wifi_connection.ssid}
Waiting for
connection...""")

        oled.centre_text(f"""{wifi_connection.ssid}
{wifi_connection.ip_addr}
Connected""")

        while True:
            self.server.check_requests()


if __name__ == "__main__":
    HttpControl().main()



# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from europi import *
from time import sleep_ms
from random import random
from europi_script import EuroPiScript
import machine

# Constant values for display
BERNOULLI_BAR_LEN = 50
BERNOULLI_BAR_WID = 6
BERNOULLI_INDI_WID = 5

# Trigger Time (some module may not be able to catch up the trigger if it is too short)
BERNOULLI_TRG_T = 20

class SingleBernoulliGate():
    def __init__(self, control_knob =  k1, control_port = ain, out_port = (cv1, cv2, cv3), visualization_para = ('P1', (0, 0), 0, -3), port3_func = 'clock', port3_source_cv = None):
        '''
        1. control_knob and control_port tunes the probability threshold
        if control_port == None, probability threshold is only determined by control_knob
        2. out_port assigns left, right, and function port accordingly
        out_port shoud have at least 2 ports (port3_func == 'none')
        3. visualization_para = (text, text_pos, tex1_pos, bar_pos_offset)
        4. port3_func and port3_source_cv determine the function of port3,
        where port3_func can be 'none', 'clock', 'and', 'or', 'xor',
        and port3_source_cv can be None or any other output port
        the output of function_port = logic(current left, source), or clock(trigger)
        '''
        self.mode_flg = 0 # mode 0: Trigger, mode 1: Gate, mode 2: Toggle
        self.coin = 0 # probability
        self.right_possibility = 0 # change every iteration
        self.left_possibility = 0
        self.right_possibility_sampled = 0 # only change when triggered
        self.left_possibility_sampled = 0

        self.port3_func = port3_func
        self.port3_source_cv = port3_source_cv

        self.control_knob = control_knob
        self.control_port = control_port

        self.left_port = out_port[0]
        self.right_port = out_port[1]
        if port3_func != 'none':
            self.function_port = out_port[2]

        self.text = visualization_para[0]
        self.text_pos = visualization_para[1]
        self.text1_pos = visualization_para[2]
        self.bar_pos_offset = visualization_para[3]

    def get_prob(self):
        if self.control_port:
            self.right_possibility = self.control_knob.percent() + self.control_port.read_voltage()/12
            self.left_possibility = 1 - self.right_possibility
        else:
            self.right_possibility = self.control_knob.percent()
            self.left_possibility = 1 - self.right_possibility

    def probability_text_visualization(self):
        oled.text(self.text + f':{self.right_possibility:.2f}', self.text_pos[0], self.text_pos[1], 1)

    def bar_visualization(self):
        # right
        oled.rect(int(OLED_WIDTH / 2),
                  int((OLED_HEIGHT - BERNOULLI_BAR_WID) / 2) + self.bar_pos_offset,
                  int(self.right_possibility * BERNOULLI_BAR_LEN),
                  BERNOULLI_BAR_WID, 1)
        # left
        oled.fill_rect(int(OLED_WIDTH / 2 - self.left_possibility * BERNOULLI_BAR_LEN),
                       int((OLED_HEIGHT - BERNOULLI_BAR_WID) / 2) + self.bar_pos_offset,
                       int(self.left_possibility * BERNOULLI_BAR_LEN),
                       BERNOULLI_BAR_WID, 1)

    def probability_sample(self):
        self.right_possibility_sampled = self.right_possibility
        self.left_possibility_sampled = self.left_possibility

    def triggered_maneuver(self):
        self.coin = random()
        if self.mode_flg == 0 or self.mode_flg == 1:
            if self.coin < (self.right_possibility_sampled):
                self.left_port.off()
                self.right_port.on()
                if self.mode_flg == 0:
                    # Draw right indicator
                    oled.fill_rect(int(OLED_WIDTH / 2 + self.right_possibility * BERNOULLI_BAR_LEN) + 2,
                                   int((OLED_HEIGHT - BERNOULLI_BAR_WID) / 2) + self.bar_pos_offset,
                                   BERNOULLI_INDI_WID,
                                   BERNOULLI_BAR_WID, 1)
            else:
                self.left_port.on()
                self.right_port.off()
                if self.mode_flg == 0:
                    # Draw left indicator
                    oled.rect(int(OLED_WIDTH / 2 - self.left_possibility * BERNOULLI_BAR_LEN) - BERNOULLI_INDI_WID - 2,
                              int((OLED_HEIGHT - BERNOULLI_BAR_WID) / 2) + self.bar_pos_offset,
                              BERNOULLI_INDI_WID,
                              BERNOULLI_BAR_WID, 1)
        else:
            if self.coin < (self.right_possibility_sampled):
                self.left_port.toggle()
                self.right_port.value(self.left_port._duty == 0)

    def function_port_maneuver(self):
        if self.port3_func == 'none':
            pass
        elif self.port3_func == 'clock':
            self.function_port.on()
        else:
            if self.port3_func == 'and':
                self.function_port.value((self.port3_source_cv._duty and self.left_port._duty) != 0)
            elif self.port3_func == 'or':
                self.function_port.value((self.port3_source_cv._duty or self.left_port._duty) != 0)
            elif self.port3_func == 'xor':
                self.function_port.value((self.port3_source_cv._duty ^ self.left_port._duty) != 0)
            else:
                self.function_port.off()

    def regular_visualization(self):
        if self.mode_flg == 0:
            oled.text('Tr', self.text1_pos, OLED_HEIGHT-8, 1)
        elif self.mode_flg == 1:
            oled.text('G', self.text1_pos, OLED_HEIGHT-8, 1)
            # Draw indicator
            if self.coin < (self.right_possibility_sampled):
                oled.fill_rect(int(OLED_WIDTH / 2 + self.right_possibility * BERNOULLI_BAR_LEN) + 2,
                               int((OLED_HEIGHT - BERNOULLI_BAR_WID) / 2) + self.bar_pos_offset,
                               BERNOULLI_INDI_WID,
                               BERNOULLI_BAR_WID, 1)
            else:
                oled.rect(int(OLED_WIDTH / 2 - self.left_possibility * BERNOULLI_BAR_LEN) - BERNOULLI_INDI_WID - 2,
                          int((OLED_HEIGHT - BERNOULLI_BAR_WID) / 2) + self.bar_pos_offset,
                          BERNOULLI_INDI_WID,
                          BERNOULLI_BAR_WID, 1)
        elif self.mode_flg == 2:
            oled.text('Tg', self.text1_pos, OLED_HEIGHT-8, 1)

    def regular_maneuver(self):
        if self.mode_flg == 0:
            self.left_port.off()
            self.right_port.off()
            if self.port3_func != 'none':
                self.function_port.off()
        elif self.mode_flg == 1:
            if self.port3_func == 'clock':
                self.function_port.off()
            else:
                pass
        elif self.mode_flg == 2:
            if self.port3_func == 'none':
                pass
            else:
                self.function_port.off()

class BernoulliGates(EuroPiScript):
    def __init__(self):
        self.toss_flg = 0
        self.first_gate = SingleBernoulliGate(control_knob =  k1,
                                              control_port = ain,
                                              out_port = (cv1, cv2, cv3),
                                              visualization_para = ('P1', (0, 0), 0, -3),
                                              port3_func = 'clock', port3_source_cv = None)
        self.second_gate = SingleBernoulliGate(control_knob =  k2,
                                               control_port = None,
                                               out_port = (cv4, cv5, cv6),
                                               visualization_para = ('P2', (int(OLED_WIDTH / 2) + 8, 0), 110, BERNOULLI_BAR_WID - 1),
                                               port3_func = 'and', port3_source_cv = cv1)

        @din.handler
        def digital_trigger():
            self.toss_flg = 1

        @b1.handler
        def mode_switch_1():
            self.first_gate.mode_flg += 1
            if self.first_gate.mode_flg == 3:
                self.first_gate.mode_flg = 0

        @b2.handler
        def mode_switch_2():
            self.second_gate.mode_flg += 1
            if self.second_gate.mode_flg == 3:
                self.second_gate.mode_flg = 0

    def main(self):
        while True:
            # Get Possibility
            self.first_gate.get_prob()
            self.second_gate.get_prob()

            # OLED Show Possibility Bar
            oled.fill(0)
            self.first_gate.probability_text_visualization()
            self.second_gate.probability_text_visualization()
            self.first_gate.bar_visualization()
            self.second_gate.bar_visualization()

            # Triggered maneuver
            if self.toss_flg == 1:
                self.first_gate.probability_sample()
                self.second_gate.probability_sample()

                self.first_gate.triggered_maneuver()
                self.second_gate.triggered_maneuver()

                self.first_gate.function_port_maneuver()
                self.second_gate.function_port_maneuver()

                self.toss_flg = 0

            # Regular maneuver
            self.first_gate.regular_visualization()
            self.second_gate.regular_visualization()
            sleep_ms(BERNOULLI_TRG_T)
            self.first_gate.regular_maneuver()
            self.second_gate.regular_maneuver()
            oled.show()

if __name__ == "__main__":
    bernoulli_gates = BernoulliGates()
    bernoulli_gates.main()



# Copyright 2023 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Sequential switch for the EuroPi

    @author Chris Iverach-Brereton <ve4cib@gmail.com>
    @date   2023-02-13
"""

from europi import *
from europi_script import EuroPiScript

from experimental.settings_menu import *
from experimental.screensaver import OledWithScreensaver

import time
import random

## Move in order 1>2>3>4>5>6>1>2>...
MODE_SEQUENTIAL=0

## Move in order 1>6>5>4>3>2>1>6>...
MODE_REVERSE=1

## Move in order 1>2>3>4>5>6>5>4>...
MODE_PINGPONG=2

## Pick a random output, which can be the same as the one we're currently using
MODE_RANDOM=3

## Instead of a traditional sequential switch, treat the outputs like a s&h shift register
#
#  CV1 will contain the latest s&h reading, with CV2-6 outputting increasingly old readings
MODE_SHIFT=4

## Use the automatic screensaver display
ssoled = OledWithScreensaver()


class SequentialSwitchDisplay(MenuItem):
    """
    A menu item that displays the current state of the sequential switch

    This isn't an editable menu item; just a visualization used as the top-level menu item.
    The actual settings are children of this item.
    """
    def __init__(self, sequential_switch, parent=None, children=None):
        super().__init__(parent=parent, children=children)

        self.sequential_switch = sequential_switch

    def draw(self, oled=ssoled):
        if self.sequential_switch.mode.value == MODE_SHIFT:
            BAR_WIDTH = OLED_WIDTH // 6
            for i in range(self.sequential_switch.num_outputs.value):
                h = max(1, int(self.sequential_switch.shift_register[i] / MAX_OUTPUT_VOLTAGE * OLED_HEIGHT))
                oled.fill_rect(BAR_WIDTH * i + 1, OLED_HEIGHT - h, BAR_WIDTH-2, h, 1)
        else:
            # Show all 6 outputs as a string on 2 lines to mirror the panel
            switches = ""
            for i in range(2):
                for j in range(3):
                    out_no = i*3 + j
                    if out_no < self.sequential_switch.num_outputs.value:
                        # output is enabled; is it hot?
                        if self.sequential_switch.current_output == out_no:
                            switches = switches + " [*] "
                        else:
                            switches = switches + " [ ] "
                    else:
                        # output is disabled; mark it with .
                        switches = switches + "  .  "

                switches = switches + "\n"

            oled.centre_text(switches, auto_show=False)


class SequentialSwitch(EuroPiScript):
    """The main workhorse of the whole module

    Copies the analog input to one of the 6 outputs, cycling which output
    whenever a trigger is received
    """

    def __init__(self):
        super().__init__()

        # The index of the current outputs
        self.current_output = 0

        # For MODE_PINGPONG, this indicates the direction of travel
        # it will always be +1 or -1
        self.direction = 1

        ## The shift register we use as s&h in MODE_SHIFT
        self.shift_register = [0] * len(cvs)

        self.visualization_dirty = True

        self.mode = SettingMenuItem(
            config_point = ChoiceConfigPoint(
                "mode",
                [
                    MODE_SEQUENTIAL,
                    MODE_REVERSE,
                    MODE_PINGPONG,
                    MODE_RANDOM,
                    MODE_SHIFT,
                ],
                MODE_SEQUENTIAL
            ),
            title="Mode",
            labels = {
                MODE_SEQUENTIAL: "Seqential",
                MODE_REVERSE: "Reverse",
                MODE_PINGPONG: "Ping-Pong",
                MODE_RANDOM: "Random",
                MODE_SHIFT: "Shift Reg.",
            },
        )

        self.num_outputs = SettingMenuItem(
            config_point = IntegerConfigPoint(
                "num_outs",
                2,
                6,
                6
            ),
            title="# Outputs",
        )

        self.menu = SettingsMenu(
            short_press_cb = lambda: ssoled.notify_user_interaction(),
            long_press_cb = lambda: ssoled.notify_user_interaction(),
            menu_items = [
                SequentialSwitchDisplay(
                    self,
                    children = [
                        self.mode,
                        self.num_outputs,
                    ]
                )
            ]
        )
        self.menu.load_defaults(self._state_filename)

        din.handler(self.on_trigger)
        b1.handler(self.on_trigger)

    def on_trigger(self):
        """Handler for the rising edge of the input clock

        Also used for manually advancing the output on a button press
        """
        # to save on clock cycles, don't use modular arithmetic
        # instead just to integer math and handle roll-over manually
        next_out = self.current_output
        if self.mode.value == MODE_SEQUENTIAL:
            next_out = next_out + 1
        elif self.mode.value == MODE_REVERSE:
            next_out = next_out - 1
        elif self.mode.value == MODE_PINGPONG:
            next_out = next_out + self.direction
        else:
            next_out = random.randint(0, self.num_outputs.value-1)

        if next_out < 0:
            if self.mode.value == MODE_REVERSE:
                next_out = self.num_outputs.value-1
            else:
                next_out = -next_out
                self.direction = -self.direction
        elif next_out >= self.num_outputs.value:
            if self.mode.value == MODE_SEQUENTIAL:
                next_out = 0
            else:
                next_out = self.num_outputs.value-2
                self.direction = -self.direction

        self.current_output = next_out

        if self.mode.value == MODE_SHIFT:
            self.shift_register.pop(-1)
            self.shift_register.insert(0, ain.read_voltage())

        self.visualization_dirty = True

    def main(self):
        """The main loop

        Connects event handlers for clock-in and button presses
        and runs the main loop
        """
        while True:
            if self.visualization_dirty or self.menu.ui_dirty:
                self.visualization_dirty = False
                ssoled.fill(0)
                self.menu.draw(ssoled)
                ssoled.show()

            if self.menu.settings_dirty:
                self.menu.save(self._state_filename)

            if self.mode.value == MODE_SHIFT:
                # CV1 gets the most-recent s&h output, all the others get older values
                for i in range(self.num_outputs.value):
                    cvs[i].voltage(self.shift_register[i])

                # turn off any unused outputs
                for i in range(self.num_outputs.value, len(cvs)):
                    cvs[i].off()
            else:
                # read the input and send it to the current output
                # all other outputs should be zero
                input_volts = ain.read_voltage()
                for i in range(len(cvs)):
                    if i == self.current_output:
                        cvs[i].voltage(input_volts)
                    else:
                        cvs[i].off()

if __name__ == "__main__":
    SequentialSwitch().main()



#!/usr/bin/env python3
# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Gaussian-based, clocked, quantized CV generator

Inspired by Magnetic Freak's Gaussian module.

@author  Chris Iverach-Brereton
@year    2024
"""

from europi import *
from europi_script import EuroPiScript

import configuration
import time

from experimental.bisect import bisect_left
from experimental.knobs import *
from experimental.random_extras import normal
from experimental.screensaver import Screensaver


class OutputBin:
    """Generic class for different output modes"""
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return self.name

    def closest(self, v):
        """Abstract function to be implemented by subclasses

        @param v  The input voltage to assign to a bin
        """
        raise Exception("Not implemented")


class ContinuousBin(OutputBin):
    """Smooth, continuous output"""
    def __init__(self, name):
        super().__init__(name)

    def closest(self, v):
        return v


class VoltageBin(OutputBin):
    """Quantizes a random voltage to the closest bin"""
    def __init__(self, name, bins):
        """Create a new set of bins

        @param name  The human-readable display name for this set of bins
        @param bins  A list of voltages we are allowed to output
        """
        super().__init__(name)
        self.bins = [float(b) for b in bins]
        self.bins.sort()

    def closest(self, v):
        """Quantize an input voltage to the closest bin. If two bins are equally close, choose the lower one.

        Our internal bins are sorted, so we can do a binary search for that sweet, sweet O(log(n)) efficiency

        @param v  A voltage in the range 0-10 to quantize
        @return   The closest voltage bin to @v
        """
        i = bisect_left(self.bins, v)
        if i == 0:
            return self.bins[0]
        if i == len(self.bins):
            return self.bins[-1]
        prev = self.bins[i - 1]
        next = self.bins[i + 1]
        if abs(v - next) < abs(v - prev):
            return next
        else:
            return prev


class DelayedOutput:
    """A class that handles setting a CV output on or after a given tick"""

    # We're not currently queued to do any output
    STATE_IDLE = 0

    # We've been assigned a time in the future to set the voltage
    STATE_WAITING = 1

    # The voltage has been applied and the gate is currently high
    STATE_GATE_HIGH = 2

    def __init__(self, cv, gate):
        """Create a new delayed output manager

        @param cv                The output for CV voltage
        @param gate              The output for a gate voltage
        """
        self.cv = cv
        self.gate = gate
        self.state = self.STATE_IDLE

        self.gate_high_tick = time.ticks_ms()
        self.gate_low_tick = time.ticks_ms()

    def process(self, now=None):
        if now is None:
            now = time.ticks_ms()

        if self.state == self.STATE_WAITING and time.ticks_diff(now, self.gate_high_tick) >= 0:
            self.cv.voltage(self.target_volts)
            self.gate.on()
            self.state = self.STATE_GATE_HIGH

        elif self.state == self.STATE_GATE_HIGH and time.ticks_diff(now, self.gate_low_tick) >= 0:
            self.gate.off()
            self.state = self.STATE_IDLE

    def voltage_at(self, v, tick, gate_duration_ms=10):
        """Specify the voltage we want to apply at the desired tick

        Call @process() to actually apply the voltage if needed

        @param v     The desired voltags (volts)
        @param tick  The tick (ms) we want the voltage to change at
        @param gate_duration_ms  The desired duration of the high cycle of the output gate
        """

        self.state = self.STATE_WAITING
        self.target_volts = v
        self.gate_high_tick = tick
        self.gate_low_tick = time.ticks_add(self.gate_high_tick, gate_duration_ms)


class Sigma(EuroPiScript):
    """The main class for this script

    Handles all I/O, renders the UI
    """

    AIN_ROUTE_NONE = 0
    AIN_ROUTE_MEAN = 1
    AIN_ROUTE_STDEV = 2
    AIN_ROUTE_JITTER = 3
    AIN_ROUTE_BIN = 4
    N_AIN_ROUTES = 5

    AIN_ROUTE_NAMES = [
        "None",
        "Mean",
        "Spread",
        "Jitter",
        "Bin"
    ]

    def __init__(self):
        super().__init__()

        self.outputs = [
            DelayedOutput(cv4, cv1),
            DelayedOutput(cv5, cv2),
            DelayedOutput(cv6, cv3)
        ]

        ## Voltage bins for bin mode
        self.voltage_bins = [
            ContinuousBin("Continuous"),
            VoltageBin("Bin 2", [0, 10]),
            VoltageBin("Bin 3", [0, 5, 10]),
            VoltageBin("Bin 6", [0, 2, 4, 6, 8, 10]),
            VoltageBin("Bin 7", [0, 1.7, 3.4, 5, 6.6, 8.3, 10]),
            VoltageBin("Bin 9", [0, 1.25, 2.5, 3.75, 5, 6.25, 7.5, 8.75, 10])
        ]

        # create bins for the quantized 1V/oct modes
        VOLTS_PER_TONE = 1.0 / 6
        VOLTS_PER_SEMITONE = 1.0 / 12
        VOLTS_PER_QUARTERTONE = 1.0 / 24
        tones = []
        semitones = []
        quartertones = []
        for oct in range(10):
            for tone in range(6):
                tones.append(oct + VOLTS_PER_TONE * tone)

            for semitone in range(12):
                semitones.append(oct + VOLTS_PER_SEMITONE * semitone)

            for quartertone in range(24):
                quartertones.append(oct + VOLTS_PER_QUARTERTONE * quartertone)

        self.voltage_bins.append(VoltageBin("Tone", tones))
        self.voltage_bins.append(VoltageBin("Semitone", semitones))
        self.voltage_bins.append(VoltageBin("Quartertone", quartertones))

        cfg = self.load_state_json()

        self.mean = cfg.get("mean", 0.5)
        self.stdev = cfg.get("stdev", 0.5)
        self.ain_route = cfg.get("ain_route", 0)
        self.voltage_bin = cfg.get("bin", 0)
        self.jitter = cfg.get("jitter", 0)

        # create the lockable knobs
        #  Note that this does mean _sometimes_ you'll need to sweep the knob all the way left/right
        #  to unlock it
        self.k1_bank = (
            KnobBank.builder(k1)
            .with_unlocked_knob("mean")
            .with_locked_knob("jitter", initial_percentage_value=cfg.get("jitter", 0.5))
            .build()
        )
        self.k2_bank = (
            KnobBank.builder(k2)
            .with_unlocked_knob("stdev")
            .with_locked_knob("bin", initial_percentage_value=self.voltage_bin / len(self.voltage_bins))
            .build()
        )

        self.config_dirty = False
        self.output_dirty = False

        self.last_interaction_at = time.ticks_ms()
        self.screensaver = Screensaver()

        self.last_clock_at = time.ticks_ms()

        self.clock_duration_ms = 0
        self.clock_duty_cycle_ms = 0

        @b1.handler
        def on_b1_rise():
            self.k1_bank.set_current("jitter")
            self.k2_bank.set_current("bin")
            self.last_interaction_at = time.ticks_ms()

        @b1.handler_falling
        def on_b1_fall():
            self.k1_bank.set_current("mean")
            self.k2_bank.set_current("stdev")
            self.config_dirty = True

        @b2.handler
        def on_b2_rise():
            self.ain_route = (self.ain_route + 1) % self.N_AIN_ROUTES
            self.config_dirty = True
            self.last_interaction_at = time.ticks_ms()

        @din.handler
        def on_din_rise():
            self.output_dirty = True
            now = time.ticks_ms()
            self.clock_duration_ms = time.ticks_diff(now, self.last_clock_at)
            self.last_clock_at = now

        @din.handler_falling
        def on_din_fall():
            now = time.ticks_ms()
            self.clock_duty_cycle_ms = time.ticks_diff(now, self.last_clock_at)

    def save(self):
        """Save the current state to the persistence file"""
        self.config_dirty = False
        cfg = {
            "ain_route": self.ain_route,
            "mean": self.mean,
            "jitter": self.jitter,
            "stdev": self.stdev,
            "bin": self.voltage_bin,
        }
        self.save_state_json(cfg)

    def read_inputs(self):
        self.mean = self.k1_bank["mean"].percent()
        self.stdev = self.k2_bank["stdev"].percent()
        self.jitter = self.k1_bank["jitter"].percent()
        self.voltage_bin = int(self.k2_bank["bin"].percent() * len(self.voltage_bins))

        # Apply attenuation to our CV-controlled input
        if self.ain_route == self.AIN_ROUTE_MEAN:
            self.mean = self.mean * ain.percent()
        elif self.ain_route == self.AIN_ROUTE_STDEV:
            self.stdev = self.stdev * ain.percent()
        elif self.ain_route == self.AIN_ROUTE_JITTER:
            self.jitter = self.jitter * ain.percent()
        elif self.ain_route == self.AIN_ROUTE_BIN:
            self.voltage_bin = int(self.k2_bank["bin"].percent() * ain.percent() * len(self.voltage_bins))

        if self.voltage_bin == len(self.voltage_bins):
            self.voltage_bin = len(self.voltage_bins) - 1  # keep the index in bounds if we reach 1.0

    def set_outputs(self, now):
        for cv in self.outputs:
            cv.process(now)

    def calculate_jitter(self, now):
        self.output_dirty = False

        for cv in self.outputs:
            if cv == self.outputs[0]:
                target_tick = now
            else:
                target_tick = time.ticks_add(now, int(abs(normal(mean = 0, stdev = self.jitter) * self.clock_duration_ms / 4)))

            x = normal(mean = self.mean * MAX_OUTPUT_VOLTAGE, stdev = self.stdev * 2)
            v = self.voltage_bins[self.voltage_bin].closest(x)
            cv.voltage_at(
                v,
                target_tick,
                self.clock_duty_cycle_ms
            )

    def main(self):
        turn_off_all_cvs()

        self.ui_dirty = True

        DISPLAY_PRECISION = 100
        prev_mean = int(self.mean * DISPLAY_PRECISION)
        prev_stdev = int(self.stdev * DISPLAY_PRECISION)
        prev_jitter = int(self.jitter * DISPLAY_PRECISION)
        prev_bin = self.voltage_bin

        while True:
            now = time.ticks_ms()

            self.read_inputs()
            if self.output_dirty:
                self.calculate_jitter(now)
            self.set_outputs(now)

            new_mean = int(self.mean * DISPLAY_PRECISION)
            new_stdev = int(self.stdev * DISPLAY_PRECISION)
            new_jitter = int(self.jitter * DISPLAY_PRECISION)

            self.ui_dirty = (self.ui_dirty or
                self.config_dirty or
                new_mean != prev_mean or
                new_stdev != prev_stdev or
                new_jitter != prev_jitter or
                self.voltage_bin != prev_bin
            )

            if self.ui_dirty:
                self.last_interaction_at = now

            prev_mean = new_mean
            prev_stdev = new_stdev
            prev_jitter = new_jitter
            prev_bin = self.voltage_bin

            if self.config_dirty:
                self.save()

            if time.ticks_diff(now, self.last_interaction_at) > self.screensaver.ACTIVATE_TIMEOUT_MS:
                self.screensaver.draw()
                last_interaction_at = time.ticks_add(now, -self.screensaver.ACTIVATE_TIMEOUT_MS*2)
            elif self.ui_dirty:
                self.ui_dirty = False
                oled.fill(0)
                oled.centre_text(f"""{self.mean:0.2f} {self.stdev:0.2f} {self.jitter:0.2f}
{self.voltage_bins[self.voltage_bin]}
CV: {self.AIN_ROUTE_NAMES[self.ain_route]}""")

if __name__ == "__main__":
    Sigma().main()



#!/usr/bin/env python3
# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Generates fixed voltages

Voltages can be configured via an optional JSON file
"""


from europi import *
from europi_script import EuroPiScript

import configuration


class OffsetVoltages(EuroPiScript):
    def __init__(self):
        super().__init__()


    @classmethod
    def config_points(cls):
        """Return the static configuration options for this class
        """
        def restrict_voltage(v):
            """If the max output voltage has been lowered, disable the too-high output
            """
            if v > europi_config.MAX_OUTPUT_VOLTAGE:
                return 0.0
            return v

        return [
            configuration.floatingPoint(
                name="CV1",
                minimum=0.0,
                maximum=europi_config.MAX_OUTPUT_VOLTAGE,
                default=restrict_voltage(0.5)
            ),
            configuration.floatingPoint(
                name="CV2",
                minimum=0.0,
                maximum=europi_config.MAX_OUTPUT_VOLTAGE,
                default=restrict_voltage(1.0)
            ),
            configuration.floatingPoint(
                name="CV3",
                minimum=0.0,
                maximum=europi_config.MAX_OUTPUT_VOLTAGE,
                default=restrict_voltage(2.0)
            ),
            configuration.floatingPoint(
                name="CV4",
                minimum=0.0,
                maximum=europi_config.MAX_OUTPUT_VOLTAGE,
                default=restrict_voltage(2.5)
            ),
            configuration.floatingPoint(
                name="CV5",
                minimum=0.0,
                maximum=europi_config.MAX_OUTPUT_VOLTAGE,
                default=restrict_voltage(5.0)
            ),
            configuration.floatingPoint(
                name="CV6",
                minimum=0.0,
                maximum=europi_config.MAX_OUTPUT_VOLTAGE,
                default=restrict_voltage(10.0)
            ),
        ]

    def main(self):
        oled.fill(0)
        oled.show()

        cv1.voltage(self.config.CV1)
        cv2.voltage(self.config.CV2)
        cv3.voltage(self.config.CV3)
        cv4.voltage(self.config.CV4)
        cv5.voltage(self.config.CV5)
        cv6.voltage(self.config.CV6)

        while True:
            pass



# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from europi import *
from time import ticks_diff, ticks_ms, sleep
from random import randint, uniform
from europi_script import EuroPiScript
import machine

'''
Probapoly
author: Nik Ansell (github.com/gamecat69)

A polyrhythmic sequencer with probability

digital_in: Clock input
analog_in: Different mode, adjusted by setting self.ainMode as follows:
- Mode 1: Analogue input toggles double time feature
- Mode 2: Analogue input voltage adjusts the upper poly value
- [default] Mode 3: Analogue input voltage adjusts the probabilities of outputs 2,3,5,6 sending gates

button_1: Short press (<500ms): Reduce pattern length (using manual pattern length is ON). Long Press (>500ms): Toggle doubletime feature
button_2: Short press (<500ms): Reduce pattern length (using manual pattern length is ON). Long Press (>500ms): Toggle Manual pattern length feature

knob_1: Select upper polyrhythm value
knob_2: Select lower polyrhythm value

output_1: Gate upper polyrhythm
output_2: Gate upper polyrhythm (50% probability)
output_3: Gate upper polyrhythm (25% probability)
output_4: Gate lower polyrhythm
output_5: Gate lower polyrhythm (50% probability)
output_6: Gate lower polyrhythm (50% probability)

'''

class Probapoly(EuroPiScript):
    def __init__(self):

        # Needed if using europi_script
        super().__init__()

        # Variables
        self.step = 1
        self.clockStep = 0
        self.resetTimeout = 2000
        self.maxPolyVal = 23
        self.upper = 1
        self.lower = 3
        self.ainValue = 0
        self.upperBernoulliProb = 50
        self.lowerBernoulliProb = 50
        self.upperProb1 = 50
        self.upperProb2 = 25
        self.lowerProb1 = 50
        self.lowerProb2 = 25
        self.doubleTime = False
        self.doubleTimeManualOverride = False
        self.manualPatternLengthFeature = False
        self.patternLength = self.lcm(self.upper, self.lower)
        self.manualPatternLength = 32  # Default manual pattern length when self.manualPatternLengthFeature is first True
        self.UPPER_BUTTON_PRESS_TIME_LIMIT = 3000 # Used as a workaround to stop phantom button presses (Issue 132)
        self.SHORT_BUTTON_PRESS_TIME_THRESHOLD = 500

        # Todo: Make this mode accessible from the UI
        # Mode 1: Analogue input toggles double time feature
        # Mode 2: Analogue input voltage adjusts the upper poly value
        # Mode 3: Analogue input voltage adjusts the probabilities of outputs 2,3,5,6 sending gates
        self.ainMode = 3

        @din.handler
        def clockRising():
            for cv in cvs:
                cv.off()
            self.handleClock()
            self.clockStep +=1
            self.step += 1

            # Reached end of pattern, or a shorter pattern is now needed (based on upper and lower values), reset step to 0
            if self.step > self.patternLength:
                self.step = 1

        @din.handler_falling
        def clockFalling():
            for cv in cvs:
                cv.off()
            if self.doubleTimeManualOverride or self.doubleTime:
                self.handleClock()
                self.clockStep +=1
                self.step += 1

                # Reached end of pattern, or a shorter pattern is now needed, reset step to 1
                if self.step > self.patternLength:
                    self.step = 1

        @b1.handler_falling
        def b1Pressed():
            if ticks_diff(ticks_ms(), b1.last_pressed()) > self.SHORT_BUTTON_PRESS_TIME_THRESHOLD and ticks_diff(ticks_ms(), b1.last_pressed()) < self.UPPER_BUTTON_PRESS_TIME_LIMIT:
                # toggle double-time feature
                self.doubleTimeManualOverride = not self.doubleTimeManualOverride
            else:
            # Short press, decrease manual pattern length if self.manualPatternLengthFeature is True
                if self.manualPatternLengthFeature:
                    self.manualPatternLength -= 1
                    self.patternLength = self.manualPatternLength

        @b2.handler_falling
        def b2Pressed():
            if ticks_diff(ticks_ms(), b2.last_pressed()) > self.SHORT_BUTTON_PRESS_TIME_THRESHOLD and ticks_diff(ticks_ms(), b2.last_pressed()) < self.UPPER_BUTTON_PRESS_TIME_LIMIT:
                # Toggle manualPatternLengthFeature
                self.manualPatternLengthFeature = not self.manualPatternLengthFeature
                if self.manualPatternLengthFeature:
                    self.patternLength = self.manualPatternLength
            else:
            # Short press, increase manual pattern length if self.manualPatternLengthFeature is True
                if self.manualPatternLengthFeature:
                    self.manualPatternLength += 1
                    self.patternLength = self.manualPatternLength

    def handleClock(self):

        # Play upper gate
        if self.step % self.upper == 0:
            cv1.value(1)

        # Output trigger with fixed and unrelated probabilities
            if randint(0,99) < self.upperProb1:
                cv2.value(1)

            if randint(0,99) < self.upperProb2:
                cv3.value(1)

        # Play lower gate
        if self.step % self.lower == 0:
            cv4.value(1)

            # Output trigger with fixed and unrelated probabilities
            if randint(0,99) < self.lowerProb1:
                cv5.value(1)

            if randint(0,99) < self.lowerProb2:
                cv6.value(1)

    # Generate pattern length by finding the lowest common multiple (LCM) and greatest common divisor (GCD)
    # https://www.programiz.com/python-programming/examples/lcm
    def lcm(self, x, y):
        return (x*y)//self.computeGcd(x,y)

    def computeGcd(self, x, y):
        while(y):
            x, y = y, x % y
        return x

    def getUpper(self):
        # Mode 2, use the analogue input voltage to set the upper ratio value
        if self.ainValue > 0.9 and self.ainMode == 2:
            self.upper = int((self.maxPolyVal / 100) * self.ainValue) + 1
        else:
            self.upper = k1.read_position(self.maxPolyVal) + 1

    def getLower(self):
        self.lower = k2.read_position(self.maxPolyVal) + 1

    def getAinValue(self):
        self.ainValue = 100 * ain.percent()
        #print(self.ainValue)

    def updateScreen(self):
        # Clear the screen
        oled.fill(0)

        rectLeftX = 20
        rectRightX = 44
        rectLength = 20

        # Calculate where the steps should be using left justification
        if self.step <= 9:
            stepLeftX = 86
        elif self.step > 9 and self.step <= 99:
            stepLeftX = 78
        else:
            stepLeftX = 70

        # current step
        oled.text(str(self.step) + '|' + str(self.patternLength), stepLeftX, 0, 1)

        # Upper + lower values
        oled.text(str(self.upper), 0, 6, 1)
        oled.rect(0 , 17, 16, 1, 1)
        oled.text(str(self.lower), 0, 22, 1)

        # probabilities
        oled.rect(rectLeftX, 6, rectLength, 8, 1)
        oled.fill_rect(rectLeftX, 6, self.upperProb1//5, 8, 1)
        oled.rect(rectRightX, 6, rectLength, 8, 1)
        oled.fill_rect(rectRightX, 6, self.upperProb2//5, 8, 1)
        oled.rect(rectLeftX, 22, rectLength, 8, 1)
        oled.fill_rect(rectLeftX, 22, self.lowerProb1//5, 8, 1)
        oled.rect(rectRightX, 22, rectLength, 8, 1)
        oled.fill_rect(rectRightX, 22, self.lowerProb2//5, 8, 1)

        if self.doubleTimeManualOverride or self.doubleTime:
            oled.text('!!', 100, 22, 1)
        if self.manualPatternLengthFeature:
            oled.text('M', 119, 22, 1)
        oled.show()

    def main(self):
        while True:
            self.getLower()
            self.getUpper()
            self.getAinValue()
            self.updateScreen()

            # Ain CV toggles doubleTime feature
            if self.ainMode == 1:
                if self.ainValue > 10:
                    self.doubleTime = True
                else:
                    self.doubleTime = False
            # Ain CV controls probability
            elif self.ainValue >= 0.9 and self.ainMode == 3:
                self.upperProb1 = int(self.ainValue * 2)
                self.upperProb2 = int(self.ainValue * 1)
                self.lowerProb1 = int(self.ainValue * 2)
                self.lowerProb2 = int(self.ainValue * 1)

            if not self.manualPatternLengthFeature:
                self.patternLength = self.lcm(self.upper, self.lower)

            # If I have been running, then stopped for longer than reset_timeout, reset the steps and clock_step to 0
            if self.clockStep != 0 and ticks_diff(ticks_ms(), din.last_triggered()) > self.resetTimeout:
                self.step = 1
                self.clockStep = 0

if __name__ == '__main__':
    dm = Probapoly()
    dm.main()



# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This is an example of how to use the SettingsMenu class.

It is not intended to be a useful script, just a convenient programming
reference to show how the menu system _can_ be used.
"""

from configuration import *
from europi import *
from europi_script import EuroPiScript
from experimental.settings_menu import *


class SettingsMenuExample(EuroPiScript):
    def __init__(self):
        super().__init__()

        on_off_labels = {
            True: "On",
            False: "Off"
        }

        # Create the main menu
        # the menu items can be generated ad-hoc like this, or passed
        # as an array generated elsewhere
        self.menu = SettingsMenu(
            menu_items = [
                SettingMenuItem(
                    config_point = FloatConfigPoint(
                        name="cv1_volts",  # every config point must have a unique name!
                        minimum=0.0,
                        maximum=europi_config.MAX_OUTPUT_VOLTAGE,
                        default=0.0,
                    ),
                    prefix="Analog",
                    title="CV1",

                    # This is a function that gets called when the user changes the setting
                    # callback_arg is an additional parameter we can pass to the callback
                    # Note: we don't NEED to use a callback; we can just read the ConfigPoint
                    # in the main loop and apply it there if desired.
                    #
                    # The callback is executed inside the button-handler interrupt routine,
                    # so any CPU-intensive or interrupt-blocking behaviour (e.g. disk IO to save
                    # a file, or rendering the display) should NOT be done via a callback function.
                    callback=self.set_analog_cv,
                    callback_arg=cv1,

                    # change this to increase the resolution of floats
                    # making it too large may cause issues with flickering
                    # generally 1-2 should be fine, 3 is usable for relatively
                    # small ranges (e.g. 0-1 with 3 decimal places)
                    float_resolution=1,

                    # For this channel only, let's also enable AIN and K1 to be used
                    # to control the voltage on CV1
                    # When the user selects "Knob" or "AIN" as the value for this channel
                    # the menu will read the value of K1/AIN and use that to choose
                    # between the manually-selectable options at a rate of 10Hz
                    autoselect_knob=True,
                    autoselect_cv=True,

                    # Create child items that can be accessed with a long-press
                    children=[
                        SettingMenuItem(
                            config_point=FloatConfigPoint(
                                name="cv2_volts",
                                minimum=0.0,
                                maximum=europi_config.MAX_OUTPUT_VOLTAGE,
                                default=0.0,
                            ),
                            prefix="Analog",
                            title="CV2",
                            callback=self.set_analog_cv,
                            callback_arg=cv2,
                            float_resolution=1,
                        ),
                        SettingMenuItem(
                            config_point=FloatConfigPoint(
                                name="cv3_volts",
                                minimum=0.0,
                                maximum=europi_config.MAX_OUTPUT_VOLTAGE,
                                default=0.0,
                            ),
                            prefix="Analog",
                            title="CV3",
                            callback=self.set_analog_cv,
                            callback_arg=cv3,
                            float_resolution=1,
                        ),
                    ]
                ),
                SettingMenuItem(
                    config_point = BooleanConfigPoint(
                        name="cv4_on",
                        default=False,
                    ),
                    prefix="Digital",
                    title="CV4",
                    callback=self.set_digital_cv,
                    callback_arg=cv4,
                    labels=on_off_labels,
                    children=[
                        SettingMenuItem(
                            config_point=BooleanConfigPoint(
                                name="cv5_on",
                                default=False,
                            ),
                            prefix="Digital",
                            title="CV5",
                            callback=self.set_digital_cv,
                            callback_arg=cv5,
                            labels=on_off_labels,
                        ),
                        SettingMenuItem(
                            config_point=BooleanConfigPoint(
                                name="cv6_on",
                                default=False,
                            ),
                            prefix="Digital",
                            title="CV6",
                            callback=self.set_digital_cv,
                            callback_arg=cv6,
                            labels=on_off_labels,
                        ),
                    ]
                ),
            ]
        )

        # Read the persistent settings file and load the settings for the menu's configuration points
        # This will trigger the callbacks as necessary
        # The _state_filename property is defined in EuroPiScript.
        self.menu.load_defaults(self._state_filename)

    def set_analog_cv(self, new_value, old_value, config_point, arg=None):
        """
        Callback function for the CV1-3 menu items

        These are floating-point settings that set voltages with 1 decimal point accuracy

        @param new_value  The new value the user has selected
        @param old_value  The previous value
        @param config_point  The ConfigPoint instance the user is editing
        @param arg  A user-defined argument, in this case cvN that we're setting the voltage for
        """
        arg.voltage(new_value)

    def set_digital_cv(self, new_value, old_value, config_point, arg=None):
        """
        Callback function for the CV4-6 menu items

        These are boolean settings that turn gate signals on/off

        @param new_value  The new value the user has selected
        @param old_value  The previous value
        @param config_point  The ConfigPoint instance the user is editing
        @param arg  A user-defined argument, in this case cvN that we're setting the voltage for
        """
        if new_value:
            arg.on()
        else:
            arg.off()

    def main(self):
        while True:
            # We can check if the menu's GUI has changed and needs to be re-drawn
            # This saves us the CPU cycles that would be used by contantly re-drawing
            # the menu. Continually re-drawing the menu CAN be done in applications
            # where preserving CPU isn't a priority
            if self.menu.ui_dirty:
                # We must clear the screen before drawing the menu and
                # call .show() afterwards
                # You may draw additional graphics/text on top of the menu
                # before calling.draw() if desired
                oled.fill(0)
                self.menu.draw()
                oled.show()

            # If any of the ConfigPoints have been changed, we can save these so they're
            # loaded automatically next time
            if self.menu.settings_dirty:
                self.menu.save(self._state_filename)


if __name__ == "__main__":
    SettingsMenuExample().main()



# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from europi import *
from europi_script import EuroPiScript

from experimental.rtc import clock

import random


class Sequence:
    """
    A single gate or CV sequence
    """
    BASE_SEQUENCE_LENGTH = 16

    def __init__(self, seed):
        self.index = 0
        self.regenerate(seed)

    def regenerate(self, seed):
        random.seed(seed)

        # randomize the length so the majority are 16, but we get some longer or shorter
        length = self.BASE_SEQUENCE_LENGTH
        r = random.random()
        if r < 0.1:
            length -= 2
        elif r < 0.25:
            length -= 1
        elif r > 0.8:
            length += 2
        elif r > 0.75:
            length += 1

        pattern = []
        for i in range(length):
            pattern.append(random.random())

        self.pattern = pattern

    def next(self):
        self.index = (self.index + 1) % len(self.pattern)

    @property
    def gate_volts(self):
        return (int(self.pattern[self.index] * 2) % 2) * europi_config.GATE_VOLTAGE

    @property
    def cv_volts(self):
        return self.pattern[self.index] * europi_config.MAX_OUTPUT_VOLTAGE


class DailyRandom(EuroPiScript):
    """
    Generates a set of pseudo-random gate and CV sequences every day

    This script requires a realtime clock. Please refer to
    experimental.clocks for supported clocks.

    If no RTC is installed & configured, the default clock will be used,
    but this will generate the same pattern every time the module is
    restarted.
    """

    SEQUENCE_LENGTH = 16

    BITMASKS = [
        0b101010101010,
        0b001100110011,
        0b000111000111,
        0b111111000000,
        0b111000111000,
        0b110011001100,
        0b010101010101,
    ]

    def __init__(self):
        super().__init__()

        self.sequences = [
            Sequence(0) for cv in cvs
        ]
        self.regenerate_sequences()

        self.trigger_recvd = False

        @din.handler
        def advance_sequence():
            for s in self.sequences:
                s.next()
            self.trigger_recvd = True

        @din.handler_falling
        def gates_off():
            for i in range(len(cvs) // 2):
                cvs[i].off()

    def regenerate_sequences(self):
        datetime = clock.localnow()
        year = datetime.year
        month = datetime.month
        day = datetime.day
        hour = datetime.hour
        minute = datetime.minute
        second = datetime.second if datetime.second is not None else 0
        weekday = datetime.weekday if datetime.weekday is not None else 1

        # bit-shift the fields around to reduce collisions
        # mask: 12 bits
        # year: 11 bits
        # month: 4 bits
        # day: 5 bits
        # hour: 6 bits
        # minute: 6 bits
        seeds = [
            self.BITMASKS[weekday - 1] ^ year ^ (month << 7) ^ day,
            self.BITMASKS[weekday - 1] ^ year ^ (month << 6) ^ day ^ ~hour,
            self.BITMASKS[weekday - 1] ^ year ^ (month << 7) ^ day ^ (hour << 6) ^ minute,
        ]

        for i in range(len(self.sequences)):
            self.sequences[i].regenerate(seeds[i % len(seeds)])

    def main(self):
        last_draw_at = clock.localnow()
        oled.centre_text(str(last_draw_at).replace(" ", "\n"))

        while True:
            now = clock.localnow()
            if now != last_draw_at:
                self.regenerate_sequences()
                oled.centre_text(str(now).replace(" ", "\n"))
                last_draw_at = now

            if self.trigger_recvd:
                self.trigger_recvd = False
                for i in range(len(self.sequences)):
                    if i < len(cvs) // 2:
                        cvs[i].voltage(self.sequences[i].gate_volts)
                    else:
                        cvs[i].voltage(self.sequences[i].cv_volts)


if __name__ == "__main__":
    DailyRandom().main()



#!/usr/bin/env python3
# Copyright 2023 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A EuroPi re-imagining of the Kompari module

Input A is provided in AIN.  K1 and K2 provide the lower and upper bounds

Outputs:
- CV1: +5V if K1 < AIN, otherwise 0
- CV2: +5V if AIN < K2, otherwise 0
- CV3: +5V if K1 < AIN < K2, otherwise 0
- CV4: max( K1, AIN )
- CV5: min( AIN, K2 )
- CV6: max( K1, min( AIN, K2 ) ) )

B1, B2, and DIN are not used

@author Chris Iverach-Brereton
@year 2023
"""

from europi import *
from europi_script import EuroPiScript

from experimental.screensaver import OledWithScreensaver

ssoled = OledWithScreensaver()

class Kompari(EuroPiScript):
    """The main Kompari script.  See module comment for usage
    """

    def __init__(self):
        super().__init__()

        @b1.handler
        def on_b1_press():
            ssoled.notify_user_interaction()

        @b2.handler
        def on_b2_press():
            ssoled.notify_user_interaction()

    @classmethod
    def display_name(cls):
        return "Kompari"

    def main(self):
        """Run the main loop
        """
        while True:
            lower_bound = k1.percent()
            upper_bound = k2.percent()
            x = ain.percent()

            if lower_bound < x:
                cv1.on()
            else:
                cv1.off()

            if x < upper_bound:
                cv2.on()
            else:
                cv2.off()

            if lower_bound < x and x < upper_bound:
                cv3.on()
            else:
                cv3.off()

            cv4.voltage(max(lower_bound, x) * MAX_OUTPUT_VOLTAGE)
            cv5.voltage(min(x, upper_bound) * MAX_OUTPUT_VOLTAGE)
            cv6.voltage(max(lower_bound, min(x, upper_bound)) * MAX_OUTPUT_VOLTAGE)

            ssoled.centre_text(f"{lower_bound:0.1f}  {x:0.1f}  {upper_bound:0.1f}", clear_first=True, auto_show=True)

if __name__ == "__main__":
    Kompari().main()



# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from europi import *
import machine
from time import ticks_diff, ticks_ms
from random import uniform
from europi_script import EuroPiScript
from europi_config import EuroPiConfig
import gc
import math
import framebuf

"""
Egressus Melodiam (Stepped Melody)
author: Nik Ansell (github.com/gamecat69)
date: 19-Feb-24
labels: clocked lfo, sequencer, CV, randomness

Known Issues:
- When the clock rate or output division changes significantly this creates temporary wave shape discontinuities
- Input clocks > 150BPM can cause waveshape glitches. Workaround this by dividing the clock first if working at 150BPM or higher.
- Some input clocks < 16ms can be occasionally missed causing glitchy waves to he output or CV pattern steps to be missed.
Workaround this issue using longer clock pulse durations or 50% duty cycle square waves.
- Performance is affected when using the controls - the poor pico has trouble keeping up with the not-yet optimized code

Possible future work:

- Med [UX]: When in unclocked mode the pattern length cannot be changed - change UI? Or maybe restrict pattern length to 1 (LFO mode only in unclocked mode?)
- Low [FEATURE]: Find a use for ain? Change unClocked clock time?
- High [PERF]: See if performance can be optimized to work with clock inputs < 16ms

Programming notes:

Outputs -> CVPattern Banks -> CV Patterns data structure:

self.cvPatternBanks[idx][self.CvPattern][step]

cvPatternBank[idx] - Output: Each list item is a reference to an output
    [maxCvPatterns] - Cv Pattern Banks: Each list item is a reference to a CV Pattern
        [maxStepLength] - Cv Patterns: Each list item is a CV value per clock step

Slew / interpolation generator structures:

CV shapes are created using interpolation formulas that fill in the gaps between CV step values in a CV pattern.
Each interpolation formula is generated by an associated slew shape function.
There are both linear and non-linear interpolation functions that create various smooth and no-so-smooth shapes.

Each slew shape function generates an array of values (samples) between two given points.
Samples are output on the CV outputs based on the list of pre-computed interpolated values (sample buffers)
A value from the sample buffer is retrieved using a Python generator function (self.slewGenerator()).
One instance of the generator function for each each slew shape exists in a list: self.slewGeneratorObjects[]

A new sample buffer (array of interpolated values) is created at each clock step.
The number of samples required in each sample buffer (one for each output) is calculated at each clock step or if
the clock rate or output division changes.

The timing and associated hanlding of input clock interrupts (clocks to din) is not perfect - python/micropython is not
known to be great at sub-second timing!
Therefore, at the end of each clock step the algorithm tunes itself to work around sample buffer overruns or underruns.
If there is a buffer underrun (not enough samples in the buffer), the previous output voltage (sample) is used until the algorithm
catches back up with itself.

slewGeneratorObjects[] contains 6 object (one for each output)
Each slewGeneratorObjects[] object is a reference to a copy of the self.slewGenerator() function.
The self.slewGenerator() function receives a buffer filled with samples and yields one sample each time it is called using next()

In order to maintain the best balance of smooth waves and Rpi pico memory usage, an algorithm is used to vary the
sample rate automatically based on the selected output division.
This causes the sample rate to be at minium when there is a slow clock and high output division.
Conversely, when there is a fast clock and low output division a higher sample rate is used to avoid unwanted steps
often caused by under-samping a wave form.

"""
# Minimum allowed time between incoming 16th note clocks. Smaller values are too fast for poor ole micropython to keep up
MIN_CLOCK_TIME_MS = 50

# Maximum allowed time between incoming 16th note clocks. This is 2 BPM, larger value causes memory issues unless the next two vals are halfed
MAX_CLOCK_TIME_MS = 3750

# Max sample rate and clock divisor to keep required buffer sizes within memory limits
MAX_SAMPLE_RATE = 32
MAX_OUTPUT_DENOMINATOR = 8

MIN_MS_BETWEEN_SAVES = 2000

# Calculate maximum sample buffer size required
SLEW_BUFFER_SIZE_IN_SAMPLES = int(
    (MAX_CLOCK_TIME_MS / 1000) * MAX_SAMPLE_RATE * MAX_OUTPUT_DENOMINATOR
)

# Reduce knob hysteresis using this value - Mutable Instruments style
KNOB_CHANGE_TOLERANCE = 0.999

# Set the maximum CV voltage using a global config value
# Important: Needs firmware v0.12.1 or higher
MAX_CV_VOLTAGE = europi_config.MAX_OUTPUT_VOLTAGE

MAX_STEP_LENGTH = 32

# Diff between incoming clocks are stored in the FiFo buffer and averaged
# Averaging over 5 values seems to deal with wonky clocks quite well
CLOCK_DIFF_BUFFER_LEN = 5

# If the clock rate changes more than this, trigger a recalculation.
# Avoids wonky waves when wonky clocks are used.
MIN_CLOCK_CHANGE_DETECTION_MS = 100

# Slightly quicker way to get integers from boolean values
BOOL_DICT = {False: 0, True: 1}

# Wave shape bit arrays
WAVE_SHAPE_IMGS = [
    bytearray(
        b"\xfe\x10\x82\x10\x82\x10\x82\x10\x82\x10\x82\x10\x82\x10\x82\x10\x82\x10\x82\x10\x82\x10\x83\xf0"
    ),  # stepUpStepDown
    bytearray(
        b"\x00\x00\x06\x00\x05\x00\t\x00\t\x00\x10\x80\x10\x80 @ @@ @ \x80\x10"
    ),  # linspace (tri)
    bytearray(b"0\x00(\x10D\x10D\x10D D\x10\x82 \x82 \x82 \x82 \x81@\x01\xc0"),  # smooth (sine)
    bytearray(
        b"\x04\x00\x04\x00\x06\x00\x06\x00\n\x00\t\x00\t\x00\x10\x80 \x80 @@ \x80\x10"
    ),  # expUpexpDown
    bytearray(
        b'\x0c\x00\x12\x00\x12\x00"\x00"\x00A\x00A\x00@\x80@\x80\x80@\x80 \x80\x10'
    ),  # sharkTooth
    bytearray(
        b"\x04\x00\x05\x00\x04\x80\x08\x80\x08@\x08@\x10 \x10 \x10   @\x10\x80\x00"
    ),  # sharkToothReverse
    bytearray(
        b"\x03\xf0\x0c\x100\x10 \x10@\x10@\x10@\x10\x80\x10\x80\x10\x80\x10\x80\x10\x80\x10"
    ),  # logUpStepDown
    bytearray(
        b"\xff\x80\x80\x80\x80\x80\x80\x80\x80\x80\x80\x80\x80\x80\x80\x80\x80@\x80@\x80 \x80\x10"
    ),  # stepUpExpDown
]


class EgressusMelodiam(EuroPiScript):
    def __init__(self):

        # Initialize variables
        self.newClockToProcess = False
        self.clockStep = 0
        self.stepPerOutput = [0, 0, 0, 0, 0, 0]
        self.nextStepPerOutput = [0, 0, 0, 0, 0, 0]
        self.CvPattern = 0
        self.resetTimeout = MAX_CLOCK_TIME_MS
        self.screenRefreshNeeded = True
        self.showNewPatternIndicator = False
        self.showNewPatternIndicatorClockStep = 0

        self.numCvPatterns = 1  # Leave at 1 due to memory limitations
        self.maxCvPatterns = 1  # Leave at 1 due to memory limitations

        self.slewArray = []
        self.lastClockTime = 0
        self.lastSlewVoltageOutputTime = [0, 0, 0, 0, 0, 0]
        self.slewGeneratorObjects = [
            self.slewGenerator([0]),
            self.slewGenerator([0]),
            self.slewGenerator([0]),
            self.slewGenerator([0]),
            self.slewGenerator([0]),
            self.slewGenerator([0]),
        ]
        self.slewShapes = [
            self.stepUpStepDown,
            self.linspace,
            self.smooth,
            self.expUpexpDown,
            self.sharkTooth,
            self.sharkToothReverse,
            self.logUpStepDown,
            self.stepUpExpDown,
        ]
        self.voltageExtremes = [0, MAX_CV_VOLTAGE]
        self.outputVoltageFlipFlops = [
            True,
            True,
            True,
            True,
            True,
            True,
        ]  # Flipflops between self.VoltageExtremes for LFO mode

        self.selectedOutput = 0
        self.lastK1Reading = 0
        self.currentK1Reading = 0
        self.lastK2Reading = 0
        self.currentK2Reading = 0

        self.running = False
        self.bufferUnderrunCounter = [0, 0, 0, 0, 0, 0]
        self.bufferOverrunSamples = [0, 0, 0, 0, 0, 0]
        self.samplesPerSec = [0, 0, 0, 0, 0, 0]
        self.msBetweenSamples = [0, 0, 0, 0, 0, 0]

        self.unClockedMode = False
        self.lastClockTime = ticks_ms()
        self.lastSaveState = ticks_ms()
        self.pendingSaveState = False
        self.previousOutputVoltage = [0, 0, 0, 0, 0, 0]
        self.slewBufferSampleNum = [0, 0, 0, 0, 0, 0]
        self.slewBufferPosition = [0, 0, 0, 0, 0, 0]
        self.bufferSampleOffsets = [0, 0, 0, 0, 0, 0]
        self.squareOutputs = [0, 0, 0, 0, 0, 0]

        self.loadState()
        # pre-create slew buffers to avoid memory allocation errors
        self.initSlewBuffers()

        # Initialize inputClockDiffs  using previous self.msBetweenClocks from loadState()
        self.inputClockDiffs = []

        # Init clock diff buffer with the default or saved value
        for n in range(CLOCK_DIFF_BUFFER_LEN):
            self.inputClockDiffs.append(self.msBetweenClocks)
        self.averageMsBetweenClocks = self.average(self.inputClockDiffs)

        # Clock rate or output division changed, recalculate optimal sample rate
        self.calculateOptimalSampleRate()

        # -----------------------------
        # Interupt Handling functions
        # -----------------------------

        @din.handler
        def clockTrigger():
            """Triggered on each rising edge into digital input. Sets running flag to true.
            Sets a flag to tell main() to process the clock step. Ignored in unclocked mode."""

            self.running = True

            if not self.unClockedMode:
                self.newClockToProcess = True

        @b1.handler_falling
        def b1Pressed():
            """Triggered when B1 is pressed and released"""
            if (
                ticks_diff(ticks_ms(), b1.last_pressed()) > 2000
                and ticks_diff(ticks_ms(), b1.last_pressed()) < 5000
            ):
                # long press generate new CV pattern
                self.generateNewRandomCVPattern(new=False, activePatternOnly=True)
                self.showNewPatternIndicator = True
                self.screenRefreshNeeded = True
                self.showNewPatternIndicatorClockStep = self.clockStep
                self.pendingSaveState = True
                #self.saveState()
            else:
                # short press change slew mode
                self.outputSlewModes[self.selectedOutput] = (
                    self.outputSlewModes[self.selectedOutput] + 1
                ) % len(self.slewShapes)
                self.screenRefreshNeeded = True
                self.pendingSaveState = True
                #self.saveState()

        @b2.handler_falling
        def b2Pressed():
            """Triggered when B2 is pressed and released"""
            if (
                ticks_diff(ticks_ms(), b2.last_pressed()) > 2000
                and ticks_diff(ticks_ms(), b2.last_pressed()) < 5000
            ):
                # long press change to unclocked mode
                self.unClockedMode = not self.unClockedMode
                if self.unClockedMode:
                    self.running = True
                self.pendingSaveState = True
                #self.saveState()

                # Update previous knob values to avoid them changing when the mode changes
                self.lastK1Reading = self.currentK1Reading
                self.lastK2Reading = self.currentK2Reading
                self.screenRefreshNeeded = True

            else:
                # short press change selected output
                self.selectedOutput = (self.selectedOutput + 1) % 6
                self.screenRefreshNeeded = True
                self.pendingSaveState = True
                #self.saveState()

    def calculateOptimalSampleRate(self):
        """Calculate optimal sample rate for smooth CV output while using minimal memory"""
        for idx in range(len(cvs)):
            self.samplesPerSec[idx] = int(
                min(2 *
                    (MAX_SAMPLE_RATE / self.outputDivisions[idx])
                    * (MAX_CLOCK_TIME_MS / self.averageMsBetweenClocks),
                    MAX_SAMPLE_RATE,
                )
            )
            self.msBetweenSamples[idx] = int(1000 / self.samplesPerSec[idx])

    def initSlewBuffers(self):
        """Create slew buffers and fill with zeros"""
        self.slewBuffers = []
        for n in range(6):  # for each output 0-5
            self.slewBuffers.append([])  # add new empty list to the buffer list
            for m in range(SLEW_BUFFER_SIZE_IN_SAMPLES):
                self.slewBuffers[n].append(0)

    def average(self, list):
        """Pythonic mean average function"""
        myList = list.copy()
        return sum(myList) / len(myList)

    def initCvPatternBanks(self):
        """Initialize CV pattern banks"""
        # Init CV pattern banks, one for each output
        self.cvPatternBanks = [[], [], [], [], [], []]
        for n in range(self.maxCvPatterns):
            self.generateNewRandomCVPattern(self)
        return self.cvPatternBanks

    def generateNewRandomCVPattern(self, new=True, activePatternOnly=False):
        """Generate new CV pattern for existing bank or create a new bank

        @param new  If true, create a new pattern/overwrite the existing one. Otherwise re-use the existing pattern
        @param activePatternOnly  If true, generate pattern for selected output. Otherwise generate pattern for all outputs

        @return True if the pattern(s) were successfully generated, otherwise False
        """
        # Note: This function is capable of working with multiple pattern banks
        #  However, due to current memory limitations only one pattern bank is used
        try:
            gc.collect()
            if new:
                # new flag provided, create new list
                if activePatternOnly:
                    self.cvPatternBanks[self.selectedOutput].append(
                        self.generateRandomPattern(MAX_STEP_LENGTH, 0, MAX_CV_VOLTAGE)
                    )
                else:
                    for pattern in self.cvPatternBanks:
                        pattern.append(
                            self.generateRandomPattern(MAX_STEP_LENGTH, 0, MAX_CV_VOLTAGE)
                        )
            else:
                # Update existing list
                if activePatternOnly:
                    self.cvPatternBanks[self.selectedOutput][self.CvPattern] = (
                        self.generateRandomPattern(MAX_STEP_LENGTH, 0, MAX_CV_VOLTAGE)
                    )
                else:
                    for pattern in self.cvPatternBanks:
                        pattern[self.CvPattern] = self.generateRandomPattern(
                            MAX_STEP_LENGTH, 0, MAX_CV_VOLTAGE
                        )
            return True
        except Exception:
            return False

    def generateRandomPattern(self, length, min, max):
        """Generate a random pattern of a desired length containing values between min and max

        @param length  The length of the desired pattern
        @param min  The minimum value for pattern elements (inclusive)
        @param max  The maximum value for pattern elements (inclusive)

        @return  The generated pattern
        """
        self.t = []
        for i in range(0, length):
            self.t.append(round(uniform(min, max), 3))
        return self.t

    def main(self):
        """Entry point - main loop. See inline comments for more info"""
        while True:
            self.updateScreen()
            self.getK1Value()
            self.getOutputDivision()

            if self.newClockToProcess:

                # Get the time difference since the last clockTime
                newDiffBetweenClocks = min(MAX_CLOCK_TIME_MS, ticks_ms() - self.lastClockTime)
                self.lastClockTime = ticks_ms()

                # Add time diff between clocks to inputClockDiffs Fifo list, skipping the first clock as we have no reference
                if self.clockStep > 0:
                    self.inputClockDiffs[self.clockStep % CLOCK_DIFF_BUFFER_LEN] = (
                        newDiffBetweenClocks
                    )

                # Clock rate change detection
                if (
                    self.clockStep >= CLOCK_DIFF_BUFFER_LEN
                    and abs(newDiffBetweenClocks - self.averageMsBetweenClocks)
                    > MIN_CLOCK_CHANGE_DETECTION_MS
                ):
                    # Update average ms between clocks
                    self.averageMsBetweenClocks = self.average(self.inputClockDiffs)
                    # Clock rate or output division changed, recalculate optimal sample rate
                    self.calculateOptimalSampleRate()

                self.handleClockStep()
                # Incremenent the clock step
                self.clockStep += 1
                self.newClockToProcess = False

            # Cycle through outputs, process when needed
            for idx in range(len(cvs)):
                if (
                    ticks_diff(ticks_ms(), self.lastSlewVoltageOutputTime[idx])
                    >= self.msBetweenSamples[idx]
                    and self.running
                ):

                    try:

                        # Do we have a sample in the buffer?
                        if self.slewBufferPosition[idx] < self.slewBufferSampleNum[idx]:
                            # Yes, we have a sample, output voltage to match the sample, reset underrun counter and advance position in buffer

                            # If a square interpolation mode (0) a precalculated value is used
                            # as the generator function for square waves was really buggy
                            if self.outputSlewModes[idx] == 0:
                                v = self.squareOutputs[idx]
                            else:
                                v = next(self.slewGeneratorObjects[idx])
                            cvs[idx].voltage(v)
                            self.previousOutputVoltage[idx] = v
                            self.bufferUnderrunCounter[idx] = 0
                        else:
                            # We do not have a sample - buffer under run
                            # Output the previous voltage to keep things as smooth as possible
                            cvs[idx].voltage(self.previousOutputVoltage[idx])
                            self.bufferUnderrunCounter[idx] += 1

                        # Advance the position in the sample/slew buffer
                        self.slewBufferPosition[idx] += 1

                        # Update the last sample output time
                        self.lastSlewVoltageOutputTime[idx] = ticks_ms()

                    except StopIteration:
                        continue

            # Save state
            if self.pendingSaveState and ticks_diff(ticks_ms(), self.lastSaveState) >= MIN_MS_BETWEEN_SAVES:
                self.saveState()
                self.pendingSaveState = False

            # If we are not being clocked, trigger a clock after the configured clock time
            if (
                self.unClockedMode
                and ticks_diff(ticks_ms(), self.lastClockTime) >= self.averageMsBetweenClocks
            ):
                self.running = True
                self.lastClockTime = ticks_ms()
                self.handleClockStep()
                self.clockStep += 1

            # If I have been running, then stopped for longer than resetTimeout, reset all steps to 0
            if (
                not self.unClockedMode
                and self.clockStep != 0
                and ticks_diff(ticks_ms(), din.last_triggered()) > self.resetTimeout
            ):
                for idx in range(len(cvs)):
                    self.stepPerOutput[idx] = 0
                # Update screen with the upcoming CV pattern
                self.screenRefreshNeeded = True
                self.pendingSaveState = True
                #self.saveState()
                self.running = False
                for cv in cvs:
                    cv.off()

                self.bufferUnderrunCounter = [0, 0, 0, 0, 0, 0]
                self.bufferOverrunSamples = [0, 0, 0, 0, 0, 0]
                self.slewBufferPosition = [0, 0, 0, 0, 0, 0]
                self.bufferSampleOffsets = [0, 0, 0, 0, 0, 0]

    def handleClockStep(self):
        """Advances step and generates new slew voltages to next value in CV pattern"""

        # Cycle through outputs and generate slew for each
        for idx in range(len(cvs)):

            # If the clockstep is a division of the output division
            if self.clockStep % (self.outputDivisions[idx]) == 0:
                # flip the flip flop value for LFO mode

                self.outputVoltageFlipFlops[idx] = not self.outputVoltageFlipFlops[idx]

                # Catch buffer over-runs by detecting that not all samples were used in the last cycle
                if self.clockStep > CLOCK_DIFF_BUFFER_LEN and not self.unClockedMode:
                    self.bufferOverrunSamples[idx] = int(
                        self.slewBufferPosition[idx] - self.slewBufferSampleNum[idx]
                    )

                    self.bufferSampleOffsets[idx] = (
                        self.bufferSampleOffsets[idx] - self.bufferOverrunSamples[idx]
                    )

                else:
                    self.bufferSampleOffsets[idx] = 0

                # Set the target number of samples for the next cycle, factoring in any previous overruns
                # Calculate the number of samples needed until the next clock
                self.slewBufferSampleNum[idx] = min(
                    SLEW_BUFFER_SIZE_IN_SAMPLES,
                    int(
                        (
                            (self.averageMsBetweenClocks / 1000)
                            * self.outputDivisions[idx]
                            * self.samplesPerSec[idx]
                        )
                        - self.bufferSampleOffsets[idx]
                    ),
                )

                # If length is one, cycle between high and low voltages (traditional LFO)
                # Each output uses a its configured slew shape
                if self.patternLength == 1:

                    # If square transition, set next output value to be one of the voltage extremes (flipping each time)
                    if self.outputSlewModes[idx] == 0:
                        self.squareOutputs[idx] = self.voltageExtremes[
                            BOOL_DICT[self.outputVoltageFlipFlops[idx]]
                        ]
                    else:
                        self.slewArray = self.slewShapes[self.outputSlewModes[idx]](
                            self.voltageExtremes[BOOL_DICT[self.outputVoltageFlipFlops[idx]]],
                            self.voltageExtremes[BOOL_DICT[not self.outputVoltageFlipFlops[idx]]],
                            self.slewBufferSampleNum[idx],
                            self.slewBuffers[idx],
                        )
                else:
                    # If square transition, just output the CV value in the pattern associated with the current step
                    if self.outputSlewModes[idx] == 0:
                        self.squareOutputs[idx] = self.cvPatternBanks[idx][self.CvPattern][
                            self.stepPerOutput[idx]
                        ]
                    else:
                        self.slewArray = self.slewShapes[self.outputSlewModes[idx]](
                            self.cvPatternBanks[idx][self.CvPattern][self.stepPerOutput[idx]],
                            self.cvPatternBanks[idx][self.CvPattern][self.nextStepPerOutput[idx]],
                            self.slewBufferSampleNum[idx],
                            self.slewBuffers[idx],
                        )

                # Update the function object reference to the generator function, passing it the latest slewArray sample buffer
                self.slewGeneratorObjects[idx] = self.slewGenerator(self.slewArray)

                # Go back to the start of the buffer
                self.slewBufferPosition[idx] = 0

                # Calculate next steps (indexs in CV patterns)
                self.stepPerOutput[idx] = ((self.stepPerOutput[idx] + 1)) % self.patternLength
                self.nextStepPerOutput[idx] = ((self.stepPerOutput[idx] + 1)) % self.patternLength

        # Hide the shreaded visual indicator after 2 clock steps
        if self.clockStep > self.showNewPatternIndicatorClockStep + 2:
            self.showNewPatternIndicator = False

    def getK1Value(self):
        """Get the k1 value, update params if changed"""

        self.currentK1Reading = k1.read_position(100) + 1

        if abs(self.currentK1Reading - self.lastK1Reading) > KNOB_CHANGE_TOLERANCE:
            # knob has moved
            if self.unClockedMode:
                # Set clock speed based on k1 value. This calc creates knob increments of 75ms
                self.averageMsBetweenClocks = (
                    self.currentK1Reading * (MAX_CLOCK_TIME_MS / MIN_CLOCK_TIME_MS) / 2
                )

                # clock rate or output division changed, calculate optimal sample rate
                self.calculateOptimalSampleRate()

            else:
                # Set pattern length
                self.patternLength = int((MAX_STEP_LENGTH / 100) * (self.currentK1Reading - 1)) + 1

            # Something changed, update screen and save state
            self.pendingSaveState = True
            #self.saveState()
            self.screenRefreshNeeded = True

        self.lastK1Reading = self.currentK1Reading

    def getOutputDivision(self):
        """Get the output division from k2"""
        self.currentK2Reading = k2.read_position(MAX_OUTPUT_DENOMINATOR) + 1

        if self.currentK2Reading != self.lastK2Reading:
            self.outputDivisions[self.selectedOutput] = k2.read_position(MAX_OUTPUT_DENOMINATOR) + 1
            self.screenRefreshNeeded = True
            self.lastK2Reading = self.currentK2Reading
            # clock rate or output division changed, calculate optimal sample rate
            self.calculateOptimalSampleRate()
            self.pendingSaveState = True
            #self.saveState()

    def saveState(self):
        """Save working vars to a save state file"""
        self.state = {
            "cvPatternBanks": self.cvPatternBanks,
            "CvPattern": self.CvPattern,
            "outputSlewModes": self.outputSlewModes,
            "outputDivisions": self.outputDivisions,
            "patternLength": self.patternLength,
            "msBetweenClocks": self.msBetweenClocks,
            "unClockedMode": self.unClockedMode,
        }
        self.save_state_json(self.state)
        self.lastSaveState = ticks_ms()

    def loadState(self):
        """Load a previously saved state, or initialize working vars, then save"""
        self.state = self.load_state_json()
        self.cvPatternBanks = self.state.get("cvPatternBanks", [])
        self.CvPattern = self.state.get("CvPattern", 0)
        self.outputSlewModes = self.state.get("outputSlewModes", [0, 0, 0, 0, 0, 0])
        self.outputDivisions = self.state.get("outputDivisions", [1, 2, 4, 1, 2, 4])
        self.patternLength = self.state.get("patternLength", 8)
        self.msBetweenClocks = self.state.get("msBetweenClocks", 976)
        self.unClockedMode = self.state.get("unClockedMode", False)

        if len(self.cvPatternBanks) == 0:
            self.initCvPatternBanks()

        self.pendingSaveState = True
        #self.saveState()
        # Let the rest of the script know how many pattern banks we have
        self.numCvPatterns = len(self.cvPatternBanks[0])

    def drawWave(self):
        """UI wave visualizations"""
        fb = framebuf.FrameBuffer(
            WAVE_SHAPE_IMGS[self.outputSlewModes[self.selectedOutput]], 12, 12, framebuf.MONO_HLSB
        )
        oled.blit(fb, 0, 20)

    def updateScreen(self):
        """Update the screen only if something has changed. oled.show() hogs the processor and causes latency."""

        # Only update if something has changed
        if not self.screenRefreshNeeded:
            return
        # Clear screen
        oled.fill(0)

        # Selected output
        oled.fill_rect(108, 0, 20, 9, 1)
        oled.text(f"{self.selectedOutput + 1}", 115, 1, 0)

        # Show division for selected output
        number = self.outputDivisions[self.selectedOutput]
        x = 111 if number >= 10 else 115
        oled.text(f"{number}", x, 12, 1)

        # Show wave for selected output
        self.drawWave()

        if self.unClockedMode:

            if self.averageMsBetweenClocks != 0:
                oled.text(f"{int(self.averageMsBetweenClocks * 2)} ms", 31, 1, 1)

        else:

            # Draw pattern length
            row1 = ""
            row2 = ""
            row3 = ""
            row4 = ""
            if self.patternLength > 24:
                # draw two rows
                row1 = "........"
                row2 = "........"
                row3 = "........"
                for i in range(24, self.patternLength):
                    row4 = row4 + "."
            elif self.patternLength > 16:
                row1 = "........"
                row2 = "........"
                for i in range(16, self.patternLength):
                    row3 = row3 + "."
            elif self.patternLength > 8:
                row1 = "........"
                for i in range(8, self.patternLength):
                    row2 = row2 + "."
            else:
                # draw one row
                for i in range(self.patternLength):
                    row1 = row1 + "."

            xStart = 27
            oled.text(row1, xStart, 0, 1)
            oled.text(row2, xStart, 6, 1)
            oled.text(row3, xStart, 12, 1)
            oled.text(row4, xStart, 18, 1)

        # Draw a visual cue for when a long button press has been detected
        # and a new random pattern is being generated
        if self.showNewPatternIndicator:
            fb = framebuf.FrameBuffer(
                bytearray(b"\x0f\x000\x80N`Q \x94\xa0\xaa\x90\xa9P\xa5@Z\x80H\x803\x00\x0c\x00"),
                12,
                12,
                framebuf.MONO_HLSB,
            )
            oled.blit(fb, 0, 0)

        oled.show()

    # -----------------------------
    # Slew functions
    # -----------------------------

    def stepUpStepDown(self, start, stop, num, buffer):
        """Produces step up, step down

        @param start  Starting value
        @param stop   Target value
        @param num    Number of samples required
        @param buffer Pointer to fill with samples

        @return  The edited buffer
        """
        c = 0
        if self.patternLength == 1:  # LFO Mode, make sure we complete a full cycle
            for i in range(num / 2):
                buffer[c] = start
                c += 1
            for i in range(num / 2):
                buffer[c] = stop
                c += 1
        else:
            for i in range(num - 1):
                buffer[c] = stop
                c += 1
        return buffer

    def linspace(self, start, stop, num, buffer):
        """Produces a linear transition

        @param start  Starting value
        @param stop   Target value
        @param num    Number of samples required
        @param buffer Pointer to fill with samples

        @return  The edited buffer
        """
        c = 0
        num = max(1, num)  # avoid divide by zero
        diff = (float(stop) - start) / (num)
        for i in range(num):
            val = (diff * i) + start
            buffer[c] = val
            c += 1
        return buffer

    def logUpStepDown(self, start, stop, num, buffer):
        """Produces a log up/step down transition

        @param start  Starting value
        @param stop   Target value
        @param num    Number of samples required
        @param buffer Pointer to fill with samples

        @return  The edited buffer
        """
        c = 0
        if self.patternLength == 1:  # LFO Mode, make sure we complete a full cycle
            for i in range(num / 2):
                i = max(i, 1)
                x = 1 - ((stop - float(start)) / (i)) + (stop - 1)
                buffer[c] = x
                c += 1
            for i in range(num / 2):
                buffer[c] = stop
                c += 1
        else:
            if stop >= start:
                for i in range(num):
                    i = max(i, 1)
                    x = 1 - ((stop - float(start)) / (i)) + (stop - 1)
                    buffer[c] = x
                    c += 1
            else:
                for i in range(num):
                    buffer[c] = stop
                    c += 1
        return buffer

    def stepUpExpDown(self, start, stop, num, buffer):
        """Produces a step up, exponential down transition

        @param start  Starting value
        @param stop   Target value
        @param num    Number of samples required
        @param buffer Pointer to fill with samples

        @return  The edited buffer
        """
        c = 0
        if stop <= start:
            for i in range(num):
                i = max(i, 1)
                x = 1 - ((stop - float(start)) / (i)) + (stop - 1)
                buffer[c] = x
                c += 1
        else:
            for i in range(num):
                buffer[c] = stop
                c += 1
        return buffer

    def smooth(self, start, stop, num, buffer):
        """Produces smooth curve using half a cosine wave

        @param start  Starting value
        @param stop   Target value
        @param num    The number of samples required
        @param buffer Pointer to fill with samples

        @return  The edited buffer
        """
        c = 0
        freqHz = 0.5  # We want to complete half a cycle
        amplitude = abs(
            (stop - start) / 2
        )  # amplitude is half of the diff between start and stop (this is peak to peak)
        if start <= stop:
            # Starting position is 90 degrees (cos) at 'start' volts
            startOffset = num
            amplitudeOffset = start
        else:
            # Starting position is 0 degrees (cos) at 'stop' volts
            startOffset = 0
            amplitudeOffset = stop
        for i in range(num):
            i += startOffset
            val = amplitude + float(
                amplitude * math.cos(2 * math.pi * freqHz * i / float(num))
            )
            buffer[c] = round(val + amplitudeOffset, 4)
            c += 1
        return buffer

    def expUpexpDown(self, start, stop, num, buffer):
        """Produces pointy exponential wave using a quarter cosine up and a quarter cosine down

        @param start  Starting value
        @param stop   Target value
        @param num    The number of samples required
        @param buffer Pointer to fill with samples

        @return  The edited buffer
        """
        c = 0
        freqHz = 0.25  # We want to complete quarter of a cycle
        amplitude = abs(
            (stop - start)
        )  # amplitude is half of the diff between start and stop (this is peak to peak)
        if start <= stop:
            startOffset = num * 2
            amplitudeOffset = start
            for i in range(num):
                i += startOffset
                val = amplitude + float(
                    amplitude * math.cos(2 * math.pi * freqHz * i / float(num))
                )
                buffer[c] = round(val + amplitudeOffset, 4)
                c += 1
        else:
            startOffset = num
            amplitudeOffset = stop
            for i in range(num):
                i += startOffset
                val = amplitude + float(
                    amplitude * math.cos(2 * math.pi * freqHz * i / float(num))
                )
                buffer[c] = round(val + amplitudeOffset, 4)
                c += 1
        return buffer

    def sharkTooth(self, start, stop, num, buffer):
        """Produces a sharktooth wave with an approximate log curve up and approximate
        exponential curve down

        @param start  Starting value
        @param stop   Target value
        @param num    The number of samples required
        @param buffer Pointer to fill with samples

        @return  The edited buffer
        """
        c = 0
        freqHz = 0.25  # We want to complete quarter of a cycle
        amplitude = abs(
            (stop - start)
        )  # amplitude is half of the diff between start and stop (this is peak to peak)
        if start <= stop:
            startOffset = num * 3
            amplitudeOffset = start - amplitude
            for i in range(num):
                i += startOffset
                val = amplitude + float(
                    amplitude * math.cos(2 * math.pi * freqHz * i / float(num))
                )
                buffer[c] = round(val + amplitudeOffset, 4)
                c += 1
        else:
            startOffset = num
            amplitudeOffset = stop
            for i in range(num):
                i += startOffset
                val = amplitude + float(
                    amplitude * math.cos(2 * math.pi * freqHz * i / float(num))
                )
                buffer[c] = round(val + amplitudeOffset, 4)
                c += 1
        return buffer

    def sharkToothReverse(self, start, stop, num, buffer):
        """Produces a reverse sharktooth wave with an approximate exponential curve up and approximate
        log curve down

        @param start  Starting value
        @param stop   Target value
        @param num    The number of samples required
        @param buffer Pointer to fill with samples

        @return  The edited buffer
        """
        c = 0
        freqHz = 0.25  # We want to complete quarter of a cycle
        amplitude = abs(
            (stop - start)
        )  # amplitude is half of the diff between start and stop (this is peak to peak)
        if start <= stop:
            startOffset = num * 2
            amplitudeOffset = start
            for i in range(num):
                i += startOffset
                val = amplitude + float(
                    amplitude * math.cos(2 * math.pi * freqHz * i / float(num))
                )
                buffer[c] = round(val + amplitudeOffset, 4)
                c += 1
        else:
            startOffset = 0
            amplitudeOffset = 1 - (amplitude - stop + 1)
            for i in range(num):
                i += startOffset
                val = amplitude + float(
                    amplitude * math.cos(2 * math.pi * freqHz * i / float(num))
                )
                buffer[c] = round(val + amplitudeOffset, 4)
                c += 1
        return buffer

    def slewGenerator(self, arr):
        """Generator function that returns the next slew sample from the specified list

        @param  The list of samples to choose from
        """
        for s in range(len(arr)):
            yield arr[s]


if __name__ == "__main__":
    dm = EgressusMelodiam()
    dm.main()



# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Binary tree based looping gate sequencer
"""

from europi import *
from europi_script import EuroPiScript

from framebuf import FrameBuffer, MONO_HLSB
from random import random as rnd


class Dcsn2(EuroPiScript):

    randomness_cv = ain
    randomness_knob = k2

    length_knob = k1

    MAX_LENGTH = 16
    MIN_LENGTH = 2

    children = [cv1, cv4]
    grandchildren = [
        [cv2, cv3],
        [cv5, cv6]
    ]

    loop_image = FrameBuffer(bytearray(b'\x1cZ\x81\x81\x81\x81Z8'), CHAR_WIDTH, CHAR_HEIGHT, MONO_HLSB)

    def __init__(self):
        super().__init__()

        self.unhandled_clock = False

        # initialize random pattern
        self.pattern = []
        for i in range(self.MAX_LENGTH):
            self.pattern.append(self.choose_random())

        def on_clock_rise():
            self.unhandled_clock = True

        def on_clock_fall():
            turn_off_all_cvs()

        def regenerate_pattern():
            for i in range(self.MAX_LENGTH):
                self.pattern[i] = self.choose_random()

        din.handler(on_clock_rise)
        din.handler_falling(on_clock_fall)
        b1.handler(on_clock_rise)
        b1.handler_falling(on_clock_fall)

        b2.handler(regenerate_pattern)

        self.set_outputs()

    def calculate_randomness(self):
        """Combine AIN & K2 to determine the probability that the pattern loops
        """
        # this will be in the range [0, 2]
        randomness = self.randomness_cv.percent() + self.randomness_knob.percent()

        # restrict to [0, 1]
        if randomness >= 1:
            randomness = 2.0 - randomness

        return randomness

    def choose_random(self):
        """
        Pick a random gate for the output pattern

        0: child 1, grandchild 1-1
        1: child 1, grandchild 1-2
        2: child 2, grandchild 2-1
        3: child 2, grandchild 2-2
        """
        return int(rnd() * 4)

    def set_outputs(self):
        turn_off_all_cvs()
        g = self.pattern[0]
        self.children[g >> 1].on()
        self.grandchildren[g >> 1][g & 1].on()

    def draw(self, pattern_length, loop_prob):
        oled.fill(0)

        active_child = self.pattern[0] >> 1
        active_grandchild = self.pattern[0] & 1

        # draw the tree with lines & circles
        oled.ellipse(OLED_WIDTH//2, 5, 4, 4, 1, True)  # root, always filled

        # children
        oled.ellipse(OLED_WIDTH//4, OLED_HEIGHT//2, 4, 4, 1, active_child == 0)
        oled.ellipse(3*OLED_WIDTH//4, OLED_HEIGHT//2, 4, 4, 1, active_child != 0)

        # grandchildren
        oled.ellipse(6, OLED_HEIGHT-5, 4, 4, 1, active_child == 0 and active_grandchild == 0)
        oled.ellipse(OLED_WIDTH//2-6, OLED_HEIGHT-5, 4, 4, 1, active_child == 0 and active_grandchild != 0)
        oled.ellipse(OLED_WIDTH//2+6, OLED_HEIGHT-5, 4, 4, 1, active_child != 0 and active_grandchild == 0)
        oled.ellipse(OLED_WIDTH-6, OLED_HEIGHT-5, 4, 4, 1, active_child != 0 and active_grandchild != 0)

        if active_child == 0:
            oled.line(OLED_WIDTH//2, 5, OLED_WIDTH//4, OLED_HEIGHT//2, 1)
            if active_grandchild == 0:
                oled.line(OLED_WIDTH//4, OLED_HEIGHT//2, 6, OLED_HEIGHT-5, 1)
            else:
                oled.line(OLED_WIDTH//4, OLED_HEIGHT//2, OLED_WIDTH//2-6, OLED_HEIGHT-5, 1)
        else:
            oled.line(OLED_WIDTH//2, 5, 3*OLED_WIDTH//4, OLED_HEIGHT//2, 1)
            if active_grandchild == 0:
                oled.line(3*OLED_WIDTH//4, OLED_HEIGHT//2, OLED_WIDTH//2+6, OLED_HEIGHT-5, 1)
            else:
                oled.line(3*OLED_WIDTH//4, OLED_HEIGHT//2, OLED_WIDTH-6, OLED_HEIGHT-5, 1)

        oled.text(f"{pattern_length}", 0, 0, 1)
        s = f"{round(loop_prob * 100)}"
        oled.blit(self.loop_image, OLED_WIDTH - CHAR_WIDTH * (len(s)+1) - 1, 0)
        oled.text(s, OLED_WIDTH - len(s) * CHAR_WIDTH, 0, 1)
        oled.text(f"{self.pattern[0]}", OLED_WIDTH//2-CHAR_WIDTH//2, OLED_HEIGHT//2 - CHAR_HEIGHT//2, 1)
        oled.show()

    def main(self):
        while True:
            r = rnd()
            loop_prob = 1.0 - self.calculate_randomness()  # 0 -> random, 1 -> loop
            pattern_length = round(self.length_knob.percent() * (self.MAX_LENGTH - self.MIN_LENGTH) + self.MIN_LENGTH)

            if self.unhandled_clock:
                self.unhandled_clock = False

                # shift the pattern over 1 step, introducing randomness as needed
                if r <= loop_prob:
                    tmp = self.pattern[pattern_length - 1]
                else:
                    tmp = self.choose_random()
                for i in range(pattern_length - 1):
                    self.pattern[pattern_length-1-i] = self.pattern[pattern_length-2-i]
                self.pattern[0] = tmp
                self.set_outputs()

            self.draw(pattern_length, loop_prob)

if __name__ == "__main__":
    Dcsn2().main()



# Copyright 2022 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from europi import *
import machine
from time import ticks_diff, ticks_ms
from europi_script import EuroPiScript
try:
    import uasyncio as asyncio
except ImportError:
    import asyncio
from random import randint

'''
Master Clock
author: Nik Ansell (github.com/gamecat69)
date: 2022-08-02
labels: clock, divider

A master clock and clock divider. Each output sends a +5V trigger/gate at different divisions of the master clock, or randomly if condigured with a division of zero.
Pulse width (gate/trigger duration) is configurable up to a maximum of 50% of the pulse width of output 1.
'''

'''
Version History (with lots missing from the early days!):
1.1 - Updates by @awoknak to reduce calls to oled.show() as it causes short hangs causes latency
1.2 - Updates by @nik:  Fix inability to activate internal/external clock with a long-press of B1
                        Remove screen1 as it is not as important now an external clock is supported
                        Fix inability to edit pulse width
                        Removed some bugs in the notes above
1.3 - Updates by @nik:  All divisions now output on step one and count from there. (e.g. /4 was: 4,8,12; now: 1,5,9)
                        Improve BPM calculations when using an external clock
'''

class MasterClockInner(EuroPiScript):
    def __init__(self):
        # Overclock the Pico for improved performance.
        machine.freq(250_000_000)
        self.step = 1
        self.clockInputNum = 1
        self.completedCycles = 0
        self.running = True
        self.resetTimeout = 3000
        self.previousStepTime = 0
        self.configMode = False
        self.k2Unlocked = False
        self.previousSelectedDivision = 0
        self.previousActiveOption = ''
        self.previousClockTime = 0
        self.inputClockDiffs = []
        self.clockSelectionScreenActive = False

        # State flag to determine if UI state has changed and display should update.
        self._updateUI = True

        self.MIN_BPM = 20  # Successfully calibrated to >= 20 and <= 240 BPM
        self.MAX_BPM = 240
        self.MIN_PULSE_WIDTH = 8
        self.MIN_AIN_VOLTAGE = 1.1
        self.MAX_DIVISION = 128
        self.MAX_PW_PERCENTAGE = 80
        self.CLOCKS_PER_QUARTER_NOTE = 4

        # Create list of available clock divisions, with 'r' (random) at the end
        self.clockDivisions = []
        for n in range(1,self.MAX_DIVISION+1):
            self.clockDivisions.append(n)
        self.clockDivisions.append('r')

        # When enabled, set msDriftCompensation to 30
        # When disabled, set msDriftCompensation to 28
        self.DEBUG = False

        # Default value is using an internal clock source
        self.externalClockInput = False

        # Set to 1 for a 4 ppqn (clock every 16th beat)
        # Set to 6 for MIDI DinSync 24 (To divide to 4 ppqn)
        # Set to 12 for MIDI DinSync 48  (To divide to 4 ppqn)
        # Note: Currently does not work well using a Din Sync input - Perhaps the pico cannot keep up?
        self.inputClockDivision = 1

        # In testing a 17ms drift was found at all tempos and a 25ms drift when editing BPM/PW
        # Adding this offset brings the BPM and pulse width back into a reasonable tolerance
        # self.msDriftCompensationConfigMode is an additional offset if configMode causes additional slowdown
        self.msDriftCompensation = 17
        self.msDriftCompensationConfigMode = 0

        # Vars to drive UI
        self.markerPositions = [ [0, 0], [69, 0], [0, 12], [40, 12], [80, 12], [0, 24], [40, 24], [80, 24]]
        self.activeOption = 1

        # Get working vars
        self.loadState()
        self.calcSleepTime()
        self.getPulseWidth()

        self.tasks = []
        for n in range(6):
            self.tasks.append(0)

        # Starts/Stops the master clock
        @b1.handler_falling
        def StartStop():
            if ticks_diff(ticks_ms(), b1.last_pressed()) > 500 and ticks_diff(ticks_ms(), b1.last_pressed()) < 4000:
                self.getClockOption()
            else:
                self.running = not self.running


        # Cycle screen and toggle config mode
        @b2.handler_falling
        def cycleScreen():
            if ticks_diff(ticks_ms(), b2.last_pressed()) > 500 and ticks_diff(ticks_ms(), b2.last_pressed()) < 4000:
                self.configMode = not self.configMode
                # This will stop the clock from running in config mode - keep here as it might be needed in the future
                #self.running = False
                if not self.configMode:
                    # config mode has just been turned off, save state and lock k2
                    self.saveState()
                    self.k2Unlocked = False
                    # Screen has changed
                    self._updateUI = True

        # Trigger clock if using an external clock, or reset if not
        @din.handler
        def dinTrigger():
            if self.externalClockInput:
                # Divide input clocks by self.inputClockDivision and trigger the clock
                if self.clockInputNum % self.inputClockDivision == 0:
                    self.clockTrigger()
                    if self.clockInputNum > 1: # Ignore the first entry as it has no reference
                        self.mSBetweenClockCycles = time.ticks_diff(ticks_ms(), self.previousClockTime)
                        self.inputClockDiffs.append(self.mSBetweenClockCycles)
                        # Only keep n values in the buffer
                        if len(self.inputClockDiffs) == 10:
                            del self.inputClockDiffs[0]

                        if self.clockInputNum > 3: # Only calculate is there are > 3 entries
                            bpm = self.calculateBpm(self.inputClockDiffs)
                            if bpm != self.bpm:
                                self.bpm = bpm
                                self._updateUI = True
                    self.previousClockTime = ticks_ms()

                self.clockInputNum += 1
            else:
                self.step = 1

    ''' Ask to use internal or external clock'''
    def getClockOption(self):
        self.clockSelectionScreenActive = True
        oled.fill(0)
        oled.text("Clock Source:", 0, 0, 1)
        oled.text("B1: Internal", 0, 9, 1)
        oled.text("B2: External", 0, 17, 1)
        self._updateUI = True
        self.updateDisplay()
        while True:
            if b1.value() == 1:
                self.externalClockInput = False
                self.running = False # Need to do this to keep it running because the b1 handler will reverse the value
                self.clockSelectionScreenActive = False
                break
            elif b2.value() == 1:
                self.externalClockInput = True
                self.clockSelectionScreenActive = False
                break
            time.sleep(0.05)

        self.saveState()
        self._updateUI = True

    def bpmFromMs(self, ms):
        return int(((1/(ms/1000))*60)/4)

    def calculateBpm(self, list):
        self.averageDiff = self.average(list)
        return self.bpmFromMs(self.averageDiff)

    def average(self, list):
        return int(sum(list) / len(list))

    '''main screen'''
    def showScreen(self):
        # k1 adjusts selected option. Remove option 1 (bpm) if using an external clock
        if self.externalClockInput:
            self.activeOption = k1.choice([2, 3, 4, 5, 6, 7, 8])
        else:
            self.activeOption = k1.choice([1, 2, 3, 4, 5, 6, 7, 8])

        oled.fill(0)
        if self.configMode and self.activeOption != 3:
            configMarker = '|'

            # if active config option changes, lock k2 and save state
            if self.previousActiveOption != self.activeOption:
                self.k2Unlocked = False
                #self.saveState()
                self._updateUI = True

            # Prevent the BPM from being configured if using an external clock input
            if self.activeOption == 1 and not self.externalClockInput:
                # read current knob value
                newBpm = self.MIN_BPM + k2.read_position(steps=(self.MAX_BPM - self.MIN_BPM + 2))
                # unlock the knob if it has reached near the same value - avoids messy UX
                if abs(newBpm - self.state.get('bpm')) <= 10:
                    self.k2Unlocked = True
                # update config value if k2 is unlocked
                if self.k2Unlocked:
                    self.bpm = newBpm
                    # calculate the new pulse width in milliseconds based on the new bpm
                    self.calcSleepTime()
                    self.getPulseWidth()
                    self._updateUI = True

            elif self.activeOption == 2:
                # read current knob value
                newPw = k2.read_position(steps=self.MAX_PW_PERCENTAGE) + 1
                # unlock the knob if it has reached near the same value - avoids messy UX
                if abs(newPw - self.state.get('pulseWidthPercent')) <= 2:
                    self.k2Unlocked = True
                # update config value if k2 is unlocked
                if self.k2Unlocked and self.pulseWidthPercent != newPw:
                    self.pulseWidthPercent = newPw
                    self.calcSleepTime()
                    self.getPulseWidth()
                self._updateUI = True

            elif self.activeOption > 2:
                # k2 adjusts clock division
                selectedDivision = k2.choice(self.clockDivisions)
                # Only adjust values if k2 has moved. This avoids a potentially annoying UX
                # self.activeOption != 3 / output 1 is disabled from configuration
                if self.previousSelectedDivision != selectedDivision and self.activeOption != 3:
                    self.outputDivisions[self.activeOption - 3] = selectedDivision

                self.previousSelectedDivision = selectedDivision
                self._updateUI = True

            self.previousActiveOption = self.activeOption

        else:
            configMarker = '.'

        oled.text(str(self.bpm) + ' bpm', 6, 0, 1)
        oled.text(str(self.pulseWidthPercent) + ':' + str(str(self.pulseWidthMs)), 75, 0, 1)
        oled.text('/' + str(self.outputDivisions[0]), 6, 12, 1)
        oled.text('/' + str(self.outputDivisions[1]), 45, 12, 1)
        oled.text('/' + str(self.outputDivisions[2]), 85, 12, 1)
        oled.text('/' + str(self.outputDivisions[3]), 6, 24, 1)
        oled.text('/' + str(self.outputDivisions[4]), 45, 24, 1)
        oled.text('/' + str(self.outputDivisions[5]), 85, 24, 1)
        oled.text(configMarker, self.markerPositions[self.activeOption-1][0], self.markerPositions[self.activeOption-1][1], 1)
        self.updateDisplay()

    ''' Holds given output (cv) high for pulseWidthMs duration '''
    async def outputPulse(self, cv):
        cv.on()
        await asyncio.sleep_ms(self.pulseWidthMs)
        cv.off()

    ''' Given a desired BPM, calculate the time to sleep between clock pulses '''
    def calcSleepTime(self):
        self.mSBetweenClockCycles = int((60000 / self.bpm / self.CLOCKS_PER_QUARTER_NOTE))

    def checkForAinBPM(self):
        val = 100 * ain.percent()
        # If there is an analogue input voltage use that for BPM. clamp ensures it is higher than MIN and lower than MAX
        if val > self.MIN_AIN_VOLTAGE:
            bpm = clamp(int((((self.MAX_BPM) / 100) * val) + self.MIN_BPM), self.MIN_BPM, self.MAX_BPM)
        else:
            # No analog input, revert to last saved state
            bpm = self.state.get("bpm", 100)

        if self.bpm != bpm:
            self.bpm = bpm
            self.calcSleepTime()
            self.getPulseWidth()
            self._updateUI = True

    def getPulseWidth(self):
        # Set max of self.MAX_PW_PERCENTAGE percent of total cycle time
        self.MAX_PULSE_WIDTH = int(self.mSBetweenClockCycles * self.MAX_PW_PERCENTAGE // 100)
        # Calc pulse width in milliseconds given the desired percentage. clamp ensures it is higher than MIN and lower than MAX
        self.pulseWidthMs = clamp((self.mSBetweenClockCycles * (self.pulseWidthPercent)//100), self.MIN_PULSE_WIDTH, self.MAX_PULSE_WIDTH)

    def computeGcd(self, x, y):
        while(y):
            x, y = y, x % y
        return x

    def lcm(self, li):
        lcm = 1
        for item in li:
            if item != 0 and item != 'r':
                lcm = lcm*item//max(self.computeGcd(lcm, item), 1)
        return lcm

    ''' Sends output pulses at required division '''
    def clockTrigger(self):

        if self.DEBUG:
            print('BPM: ' + str(self.bpm) + ' cycle: ' + str(self.mSBetweenClockCycles) + ' PW:' + str(self.pulseWidthMs))

        for idx, output in enumerate(self.outputDivisions):
            if output != 'r':
                if (self.step - 1) % output == 0:
                    if self.tasks[idx] != 0 and not self.tasks[idx].done() and self.DEBUG:
                        print(f'Task: {idx} is not done')
                    if self.DEBUG:
                        print(f'calling outputPulse({idx}) on division {output}')
                    self.tasks[idx] = asyncio.create_task(self.outputPulse(cvs[idx]))
            else:
                # Fire pulses randomly
                if randint(0, 1):
                    self.tasks[idx] = asyncio.create_task(self.outputPulse(cvs[idx]))

        # advance/reset clock step, resetting at the lowest common multiple
        if self.step < self.lcm(self.outputDivisions):
            self.step += 1
        else:
            self.completedCycles += 1
            self.step = 1

        # Get time of last step to use in the auto reset function
        self.previousStepTime = ticks_ms()

        # Display may update on clock trigger.
        self._updateUI = True

        # Debug task output to check for overrunning tasks i.e. memory leaks
        if self.DEBUG:
            for i in self.tasks:
                if i != 0:
                    print(f'[{i}] done: {str(i.done())}. state: {str(i.state)}. data: {str(i.data)}. coro: {str(i.coro)}')
                else:
                    print(0)

    def updateDisplay(self):
        """Update the display if UI state has changed."""
        if self._updateUI:
            oled.show()
            self._updateUI = False

    ''' Save working vars to a save state file'''
    def saveState(self):
        self.state = {
            "bpm": self.bpm,
            "pulseWidthPercent": self.pulseWidthPercent,
            "externalClockInput": self.externalClockInput,
            "outputDivisions": self.outputDivisions
        }
        self.save_state_json(self.state)
        if self.DEBUG:
            print('State saved')

    ''' Load a previously saved state, or initialize working vars, then save'''
    def loadState(self):

        self.state = self.load_state_json()

        self.bpm = self.state.get("bpm", 100)
        self.pulseWidthPercent = self.state.get("pulseWidthPercent", 50)
        self.externalClockInput = self.state.get("externalClockInput", False)
        self.outputDivisions = self.state.get("outputDivisions", [1,2,4,8,16,32])

        self.saveState()

    async def main(self):
        while True:
            if not self.clockSelectionScreenActive:
                self.showScreen()

            # Auto reset function after resetTimeout
            if self.step != 0 and ticks_diff(ticks_ms(), self.previousStepTime) > self.resetTimeout:
                 self.step = 1
                 self.completedCycles = 0

            if not self.configMode and not self.externalClockInput:
                self.checkForAinBPM()

            if self.running and not self.externalClockInput:
                self.clockTrigger()
                self.calcSleepTime()
                if self.configMode:
                    await asyncio.sleep_ms(int(self.mSBetweenClockCycles - self.msDriftCompensation - self.msDriftCompensationConfigMode))
                else:
                    await asyncio.sleep_ms(int(self.mSBetweenClockCycles - self.msDriftCompensation))
            else:
                # need to add this otherwise the async tasks never start
                await asyncio.sleep_ms(0)

class MasterClock(EuroPiScript):
    def __init__(self):
        pass
    def main(self):
        mc = MasterClockInner()
        el = asyncio.get_event_loop()
        el.create_task(mc.main())
        el.run_forever()

if __name__ == '__main__':
    m = MasterClock()
    m.main()
